""" The module that stores all the default transformers and the register functions.
"""
import re
import math
import sys
import json
import inspect
import logging
import pathlib
import importlib
from abc import abstractmethod

import numpy as np
import pandas as pd
import pandera.pandas as pa


from . import errormanager
from. import exceptions
from . import validate
from . import make_value
from . import make_labels
from . import base
from . import loader

logger = logging.getLogger("ontoweaver")


def register(transformer_class):
    """Adds the given transformer class to those available to OntoWeaver.

    The given class should inherit from ontoweaver.base.Transformer

    Example::

        import ontoweaver

        class user_transformer(ontoweaver.base.Transformer):
            def __init__(self, target, target_element_properties, edge=None, columns=None, **kwargs):
                super().__init__(target, target_element_properties, edge, columns, **kwargs)

            def __call__(self, row, i):
                for key in self.columns:
                    yield str(row[key])

        ontoweaver.transformer.register( user_transformer )

        # The mapping can now use "user_transformer" in the transformers list.

    Args:
        transformer_class: The class to add to the ontoweaver.transformer module.
    """

    if not issubclass(transformer_class, base.Transformer):
        logging.error(f"{transformer_class.__name__} should inherit from ontoweaver.base.Transformer.", section="transformer.register", exception = exceptions.InterfaceInheritanceError)
    current = sys.modules[__name__]
    logging.debug(f"Adding transformer {transformer_class.__name__}")
    setattr(current, transformer_class.__name__, transformer_class)


# NOTE: transformers pass all kwargs to superclass to allow it to show
#       the (additional) user-defined arguments when calling __repr__.


def import_from_path(file_path):
    """Import the given Python file path as a module."""
    # See https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
    module_name = pathlib.Path(file_path).stem
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    assert spec
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def register_all(module_path):
    for mpath in module_path:
        logger.info(f"Look for transformers in `{mpath}`")
        mod = import_from_path(mpath)
        for name,cls in mod.__dict__.items():
            if inspect.isclass(cls):
                logger.debug(f"{cls}")
                if issubclass(cls, base.Transformer):
                    logger.info(f"    Register transformer: `{cls}`")
                    register(cls)


class map(base.Transformer):
    """Transformer subclass used for the simple mapping of cell values of defined columns and creating
    nodes with their respective values as id."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True):
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            for key in columns:
                if key not in row:
                    self.error(f"Column '{key}' not found in data", section="map.call",
                               exception=exceptions.TransformerDataError)
                else:
                    yield row[key]

    def __init__(self,
            properties_of,
            label_maker = None,
            branching_properties = None,
            columns=None,
            output_validator: validate.OutputValidator = None,
            multi_type_dict = None,
            raise_errors = True,
            **kwargs
        ):
        """
        Initialize the map transformer.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """

        self.value_maker = self.ValueMaker(raise_errors=raise_errors)

        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )

        if not self.columns:
            self.error(f"No column declared for the `{type(self).__name__}` transformer, did you forgot to add a `columns` keyword?", section="map.call", exception = exceptions.TransformerInputError)


    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The cell value if valid.

        Raises:
            Warning: If the cell value is invalid.
        """
        for item in super().__call__(row, i):
            yield item


class split(base.Transformer):
    """Transformer subclass used to split cell values at defined separator and label_maker nodes with
    their respective values as id."""

    class ValueMaker(make_value.ValueMaker):

        def __init__(self, raise_errors: bool = True, separator: str = None):
            self.separator = separator
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            for key in columns:
                if key not in row:
                    self.error(f"Column `{key}` not found in row:\n{row}",
                        "transformer.split",
                        exception = exceptions.TransformerConfigError)

                val = row[key]
                if isinstance(val, str):
                    items = val.split(self.separator)
                    for item in items:
                        yield item

                elif not base.is_not_null(val):
                    logger.debug(f"Value is null, I'll let my caller skip it.")
                    yield val  # Will be passed by super.__call__

                else:
                    try:  # Try generic access.
                        for item in val:
                            yield item
                    except Exception as e:
                        self.error(f"Cannot skip or iterate over {type(val)}: `{val}`. {e}",
                            "transformer.split",
                            exception = exceptions.TransformerDataError)

    def __init__(self,
        properties_of,
        label_maker = None,
        branching_properties = None,
        columns=None,
        output_validator: validate.OutputValidator = None,
        multi_type_dict = None,
        raise_errors = True,
        separator = None,
        **kwargs
    ):
        """
        Initialize the split transformer.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            sep: Character(s) to use for splitting.
            output_validator: the OutputValidator object used for validating transformer output.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
            separator: The character(s) to use for splitting the cell values. Defaults to ",".
        """

        assert columns, "I need at least 1 column to operate."
        assert isinstance(columns, list), "I need at least 1 column to operate."
        assert len(columns) >= 1, "I need at least 1 column to operate."

        self.separator = separator

        self.value_maker = self.ValueMaker(raise_errors=raise_errors, separator=self.separator)

        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict = multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )


class cat(base.Transformer):
    """Transformer subclass used to concatenate cell values of defined columns and label_maker nodes with
    their respective values as id."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True):
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            formatted_items = ""
            for key in columns:
                formatted_items += str(row[key])

            yield formatted_items

    def __init__(self,
            properties_of,
            label_maker = None,
            branching_properties = None,
            columns=None,
            output_validator: validate.OutputValidator = None,
            multi_type_dict = None,
            raise_errors = True,
            **kwargs
        ):
        """
        Initialize the cat transformer.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            output_validator: the OutputValidator object used for validating transformer output.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
        """

        self.value_maker = self.ValueMaker(raise_errors=raise_errors)

        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )

    def __call__(self, row, i):
        """
        Process a row and yield concatenated items as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The concatenated string from the cell values.
        """
        if not self.columns:
            self.error(f"No column declared for the {type(self).__name__} transformer, did you forgot to add a `columns` keyword?", section=f"{type(self).__name__}.call", exception = exceptions.TransformerInputError)

        for item in super().__call__(row, i):
            yield item


class cat_format(base.Transformer):
    """Transformer subclass used to concatenate cell values of defined columns and label_maker nodes with
    their respective values as id."""

    class ValueMaker(make_value.ValueMaker):

        def __init__(self, raise_errors: bool = True, format_string: str = None):
            self.format_string = format_string
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):

            formatted_string = self.format_string.format_map(row)
            yield formatted_string

    def __init__(self,
            properties_of,
            label_maker = None,
            branching_properties = None,
            columns=None,
            output_validator: validate.OutputValidator = None,
            multi_type_dict = None,
            raise_errors = True,
            format_string = None,
             **kwargs
         ):
        """
        Initialize the cat_format transformer.

        Args:.
            target_element_properties: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            format_string: A format string containing the column names to assemble.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """

        self.value_maker = self.ValueMaker(
            raise_errors=raise_errors,
            format_string=format_string
        )

        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )

        if not format_string:  # Neither empty string nor None.
            self.error(f"The `format_string` parameter of the `{type(self).__name__}` transformer cannot be an empty string.")



class rowIndex(base.Transformer):
    """Transformer subclass used for the simple mapping of nodes with row index values as id."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True):
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            yield i

    def __init__(self,
            properties_of,
            label_maker = None,
            branching_properties = None,
            columns=None,
            output_validator: validate.OutputValidator = None,
            multi_type_dict = None,
            raise_errors = True,
            **kwargs
        ):
        """
        Initialize the rowIndex transformer.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """

        self.value_maker = self.ValueMaker(raise_errors=raise_errors)

        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )


class nested(base.Transformer):
    """Transformer subclass used for accessing a value within
       nested dictionaries or dataframes."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True):
            super().__init__(raise_errors)

        def __call__(self, keys, dic, i):
            value = self.nested(keys, dic, i)
            if value:
                return [value]
            else:
                return []

        def nested(self, keys, dic, i, depth = " "):
            depth += "| "
            logger.debug(f"{depth}Received: {type(dic)}{keys}:\n{dic}")

            if dic is None:
                return None

            if isinstance(dic, str) and not keys:
                # Break recursivity.
                logger.debug(f"{depth}Ended with: `{dic}`")
                return dic

            if not isinstance(keys, str):
                logger.debug(f"{depth}I can iterate over keys")
                if isinstance(dic, str):
                    if ':' in dic and '{' in dic and '}' in dic:
                        # Probably a Python/JSON dictionary.
                        logger.debug(f"{depth}I'm parsing a JSON dictionary string.")
                        dic = json.loads(dic)
                        logger.debug(f"{depth}{dic}")
                    assert not isinstance(dic, str)

                # Consider it an object with bracket access, we just pass it.
                if isinstance(dic, np.ndarray):
                    if dic.shape == (0,):
                        return None
                    else:
                        return self.nested( keys[1:], dic[keys[0]], i, depth )

                elif keys[0] not in dic:
                    if isinstance(dic, dict):
                        available_keys = ', '.join(dic.keys())
                    elif isinstance(dic, pd.DataFrame):
                        available_keys = ', '.join(dic.columns)
                    else:
                        available_keys = f"[unknown object type `{type(dic)}`, I cannot read its keys or it does not have ones]"

                    msg = f"Key '{keys[0]}' not found in data, I can only see keys: {available_keys}."
                    msg += f" Object repr: {dic}."
                    self.error(msg, section="nested.call",
                               exception=exceptions.TransformerDataError)
                else:
                    # Recursive call until exhaustion.
                    logger.debug(f"{depth}Get: {type(dic)}[{keys[0]}]")
                    return self.nested( keys[1:], dic[keys[0]], i, depth )

    def __init__(self,
        properties_of,
        label_maker = None,
        branching_properties = None,
        columns=None,
        output_validator: validate.OutputValidator = None,
        multi_type_dict = None,
        raise_errors = True,
        **kwargs
    ):

        assert columns, "I need at least 1 key to operate."
        assert isinstance(columns, list), "I need at least 1 key to operate."
        assert len(columns) >= 1, "I need at least 1 key to operate."

        self.value_maker = self.ValueMaker(raise_errors=raise_errors)
        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )

        self.keys = columns
        logger.debug(f"keys: {self.keys}")
        if not self.keys:
            self.error(f"No key declared for the `{type(self).__name__}` transformer, did you forgot to add a `keys` keyword?", section="nested.call", exception = exceptions.TransformerInputError)

    def __call__(self, row, i):
        for item in super().__call__(row, i):
            yield item


class split_nested(base.Transformer):

    def __init__(self,
        properties_of,
        label_maker = None,
        branching_properties = None,
        columns=None,
        output_validator: validate.OutputValidator = None,
        multi_type_dict = None,
        raise_errors = True,
        separator = None,
        **kwargs
    ):
        """
        FIXME doc
        """

        logger.debug(f"COLUMNS: {type(columns)}\n{columns}")
        assert columns, "I need 2 keys to operate."
        assert isinstance(columns, list), "I need several keys."
        assert len(columns) >= 2, "I need 2 keys, or you should use either split or nested."

        self.split = split(
            properties_of,
            label_maker,
            branching_properties,
            [columns[0]],
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            separator = separator,
            **kwargs,
        )

        keys = columns[1:]
        if not isinstance(keys, list):
            keys = [keys]

        self.nested = nested(
            properties_of,
            label_maker,
            branching_properties,
            keys,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs,
        )

        super().__init__(properties_of,
            self.split.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )

    def __call__(self, row, i):
        for rowval in self.split.value_maker(self.split.columns, row, i):
            val = self.nested.value_maker(self.nested.keys, rowval, i)

            value, edge_type, node_type, reverse_edge = self.create(val, row)
            if base.is_not_null(value):
                yield value, edge_type, node_type, reverse_edge


class capitalize(map):
    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs,
        with first letter in uppercase.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The capitalized cell value if valid.

        Raises:
            Warning: If the cell value is invalid.
        """
        for item in super().__call__(row, i):
            value, edge_type, node_type, reverse_edge = item
            val = value.capitalize()
            yield val, edge_type, node_type, reverse_edge


class lower(map):
    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs,
        with all letters in lowercase.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The capitalized cell value if valid.

        Raises:
            Warning: If the cell value is invalid.
        """
        for item in super().__call__(row, i):
            value, edge_type, node_type, reverse_edge = item
            val = value.lower()
            yield val, edge_type, node_type, reverse_edge


class upper(map):
    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs,
        with all letters in uppercase.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The capitalized cell value if valid.

        Raises:
            Warning: If the cell value is invalid.
        """
        for item in super().__call__(row, i):
            value, edge_type, node_type, reverse_edge = item
            val = value.lower()
            yield val, edge_type, node_type, reverse_edge


class lower_capitalize(map):
    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs,
        with first letter in uppercase, and all others in lowercase.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The capitalized cell value if valid.

        Raises:
            Warning: If the cell value is invalid.
        """
        for item in super().__call__(row, i):
            value, edge_type, node_type, reverse_edge = item
            val = value.lower().capitalize()
            yield val, edge_type, node_type, reverse_edge


class translate(base.Transformer):
    """Translate the targeted cell value using a tabular mapping and yield a node with using the translated ID."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, translate, translate_from, translate_to, raise_errors: bool = True):
            self.translate = translate
            self.translate_from = translate_from
            self.translate_to = translate_to
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):

            for key in columns:
                if key not in row:
                    self.error(f"Column '{key}' not found in data", section="translate", 
                               exception = exceptions.TransformerDataError)
                cell = row[key]
                if cell in self.translate:
                    yield self.translate[cell]
                else:
                    logger.warning(f"Row {i} does not contain something to be translated from `{self.translate_from}` to `{self.translate_to}` at column `{key}`.")

    def __init__(self,
            properties_of,
            label_maker = None,
            branching_properties = None,
            columns=None,
            output_validator: validate.OutputValidator = None,
            multi_type_dict = None,
            raise_errors = True,
            **kwargs
        ):
        """
        Constructor.

        NOTE: The user should provide at least either `translations` or `translations_file`, but not both.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            translations: A dictionary figuring what to replace (keys) with which string (values).
            translations_file: A filename pointing to a tabular file readable by Pandas' csv_read.
            translate_from: The column in the file containing what to replace.
            translate_to: The column in the file containing the replacement string.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
            kwargs: Additional arguments to pass to a Loader function (e.g. if you want to load TSVs, "sep=TAB", reads the translations_file as tab-separated).
        """

        self.map = map(properties_of, label_maker, branching_properties, columns, output_validator, multi_type_dict, **kwargs)

        lpf = loader.LoadPandasFile()
        lpd = loader.LoadPandasDataframe()
        lrf = loader.LoadOWLFile()
        lrg = loader.LoadOWLGraph()

        # Since we cannot expand kwargs, let's recover what we have inside.
        self.translations = kwargs.get("translations", None)
        self.translations_file = kwargs.get("translations_file", None)
        self.translate_from = kwargs.get("translate_from", None)
        self.translate_to = kwargs.get("translate_to", None)

        if self.translations and self.translations_file:
            self.error(f"Cannot have both `translations` (=`{self.translations}`) and `translations_file` (=`{self.translations_file}`) defined in a {type(self).__name__} transformer.", section="translate", exception = exceptions.TransformerInterfaceError)

        if self.translations:
            self.translate = self.translations
            logger.debug(f"\t\t\tManual translations: `{self.translate}`")
        elif self.translations_file:
            logger.debug(f"\t\t\tGet translations from file: `{self.translations_file}`")
            if not self.translate_from:
                self.error(f"No translation source column declared for the `{type(self).__name__}` transformer using translations_file=`{self.translations_file}`, did you forget to add a `translate_from` keyword?", section="translate.init", exception = exceptions.TransformerInterfaceError)
            if not self.translate_to:
                self.error(f"No translation target column declared for the `{type(self).__name__}` transformer using translations_file=`{self.translations_file}`, did you forget to add a `translate_to` keyword?", section="translate.init", exception = exceptions.TransformerInterfaceError)
            else:
                # self.translations_file = translations_file
                # self.translate_from = translate_from
                # self.translate_to = translate_to

                # Possible arguments from the `translate` section.
                mapping_args = ["translations", "translations_file", "translate_from", "translate_to"]
                # Possible Python attributes.
                mapping_args += ["subclass"]
                # Discard match
                mapping_args += ["match"]
                # All possible arguments found in a YAML mapping.
                for attr in dir(base.MappingParser):
                    if re.match("^k_", attr):
                        mapping_args += getattr(base.MappingParser, attr)

                # Keep only the user-passed arguments that are not in possible YAML keywords.
                more_args = {k:v for k,v in kwargs.items() if k not in mapping_args}
                if more_args['sep'] == 'TAB': # FIXME why the fuck is this changed somehow?
                    more_args['sep'] = '\t'

                logger.debug(f"\t\t\tAdditional user-passed arguments for the load function: {more_args}")

                self.df = pd.DataFrame()
                for with_loader in [lpf, lpd, lrf, lrg]:
                    if with_loader.allows([self.translations_file]):
                        logger.debug(f"\t\t\tUsing loader: {type(with_loader).__name__}")
                        self.df = with_loader.load([self.translations_file], **more_args)
                        break

                if self.df.empty:
                    self.error(f"I was not able to load a valid translations_file from: `{self.translations_file}`")

                logger.debug(f"Loaded a DataFrame: {self.df}")

                if self.translate_from not in self.df.columns:
                    self.error(f"Source column `{self.translate_from}` not found in {type(self).__name__} transformer’s translations file `{self.translations_file}`, available headers: `{','.join(self.df.columns)}`.", section="translate.init", exception = exceptions.TransformerDataError)

                if self.translate_to not in self.df.columns:
                    self.error(f"Target column `{self.translate_to}` not found in {type(self).__name__} transformer’s translations file `{self.translations_file}`, available headers: `{','.join(self.df.columns)}`.", section="translate.init", exception = exceptions.TransformerDataError)

                self.translate = {}
                for i,row in self.df.iterrows():
                    frm = row[self.translate_from]
                    to = row[self.translate_to]
                    if frm in self.translate and self.translate[frm] != to:
                        logger.warning(f"The key `{frm}` already exists in the translation table, and translated to `{self.translate[frm]}`. It now translates to `{to}`. You may want to avoid such duplicates in translation tables.")
                    if frm and to:
                        self.translate[frm] = to
                    else:
                        logger.warning(f"Cannot translate frm `{self.translate_from}` to `{self.translate_to}`, invalid translations values at row {i} of file `{self.translations_file}`: `{frm}` => `{to}`. I will ignore this translation.")

        else:
            self.error(f"When using a {type(self).__name__} transformer, you must define either `translations` or `translations_file`.", section="translate.init", exception = exceptions.TransformerInterfaceError)


        if not self.translate:
            self.error("No translation found, did you forget the `translations` keyword?", section="translate.init", exception = exceptions.TransformerInterfaceError)

        self.value_maker = self.ValueMaker(
            self.translate,
            self.translate_from,
            self.translate_to,
            raise_errors=raise_errors
        )

        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )
        
    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The cell value if valid.

        Raises:
            Warning: If the cell value or the translation is invalid.
        """
        if not self.columns:
            self.error(f"No column declared for the {type(self).__name__} transformer, did you forgot to add a `columns` keyword?", section="translate", exception = exceptions.TransformerDataError)

        for item in super().__call__(row, i):
            yield item


class string(base.Transformer):
    """A transformer that makes up the given static string instead of extractsing something from the table."""

    class ValueMaker(make_value.ValueMaker):
        def __init__(self, raise_errors: bool = True, string: str = None):
            self.string = string
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            yield self.string

    def __init__(self,
            properties_of,
            label_maker = None,
            branching_properties = None,
            columns=None,
            output_validator: validate.OutputValidator = None,
            multi_type_dict = None,
            raise_errors = True,
            **kwargs
        ):
        """
        Constructor.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            value: The string to use.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """

        self.value = kwargs.get("value", None)
        self.value_maker = self.ValueMaker(raise_errors=raise_errors, string=self.value)

        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )

    def __call__(self, row, i):
        """
        Process a row and yield cell values as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The cell value if valid.

        Raises:
            Warning: If the cell value is invalid.
        """
        if not self.value:
            self.error(f"No value passed to the {type(self).__name__} transformer, did you forgot to add a `value` keyword?", section="string.call", exception = exceptions.TransformerInterfaceError)

        for item in super().__call__(row, i):
            yield item


class replace(base.Transformer):
    """Transformer subclass used to remove characters that are not allowed from cell values of defined columns.
     The forbidden characters are defined by a regular expression pattern, and are substituted with a user-defined
     character or removed entirely. In case the cell value is made up of only forbidden characters, the node is not
     created and a warning is logged."""

    class ValueMaker(make_value.ValueMaker):

        def __init__(self, raise_errors: bool = True, forbidden = None, substitute = None):

            self.forbidden = forbidden
            self.substitute = substitute
            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            for key in columns:

                if not base.is_not_null(row[key]):
                    yield row[key]
                else: 
                    logger.debug(
                        f"Setting forbidden characters: {self.forbidden} for `replace` transformer, with substitute character: `{self.substitute}`.")
                    formatted = re.sub(self.forbidden, self.substitute, row[key])

                    strip_formatted = formatted.strip(self.substitute)
                    logger.debug(f"Formatted value: {strip_formatted}")
                    yield strip_formatted

    def __init__(self,
            properties_of,
            label_maker = None,
            branching_properties = None,
            columns=None,
            output_validator: validate.OutputValidator = None,
            multi_type_dict = None,
            raise_errors = True,
            **kwargs
        ):
        """
        Constructor.

        Args:
            properties_of: Properties of the node.
            value_maker: the ValueMaker object used for the logic of cell value selection for each transformer.
            label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
            branching_properties: in case of branching on cell values, the dictionary holding the properties for each branch.
            columns: The columns to be processed.
            forbidden: The regular expression pattern to match forbidden characters.
            substitute: The string to replace forbidden characters with.
            output_validator: the OutputValidator object used for validating transformer output.
            multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
            raise_errors: if True, the caller is asking for raising exceptions when an error occurs
        """
        self.forbidden = kwargs.get("forbidden", r'[^a-zA-Z0-9_`.()]') # By default, allow alphanumeric characters (A-Z, a-z, 0-9),
        # underscore (_), backtick (`), dot (.), and parentheses (). TODO: Add or remove rules as needed based on errors in Neo4j import.
        self.substitute = kwargs.get("substitute", "")

        self.value_maker = self.ValueMaker(raise_errors=raise_errors, forbidden=self.forbidden, substitute=self.substitute)

        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )


class boolean(base.Transformer):
    """A transformer that can map any set of values onto a boolean pair.

    It considers a set of truth values, along with a set of falsehood values,
    and then set the node ID to the user's true or false value.

    If no configuration is given for ``consider_true`` and ``consider_false``,
    OntoWeaver will use Python's `bool(value)`` to assert the truth of the value
    passed from the cell.

    If ``output_true`` or ``output_false`` are omitted, they will default to "True"
    and "False".

    For instance:

    .. code:: yaml

        - boolean:
            column: my_column
            via_relation: my_relation
            consider_true:
                - Y
                - Yes
                - yes
            output_true: my_truth
            consider_false:
                - N
                - No
                - no
            output_false: my_falsehood

    Is equivalent to:

    .. code:: python

        if value in ["Y", "Yes", "yes"]:
            yield "my_truth"
        if value in ["N", "No", no"]:
            yield "my_falsehood"

    """

    class ValueMaker(make_value.ValueMaker):

        def __init__(self, raise_errors: bool = True, output_true = "True", output_false = "False", consider_true = None, consider_false = None):
            if (not consider_true and consider_false) or (consider_true and not consider_false):
                msg = "I can only handle both `consider_true` and `consider_false` being `None` at the same time. But here, one of them is `None` and the other is not."
                logger.error(msg)
                raise exceptions.TransformerConfigError(msg)

            elif consider_true and consider_false:
                if not isinstance(consider_true, list) or not isinstance(consider_false, list):
                    msg = "I can only consider both `consider_true` and `consider_false` being lists."
                    logger.error(msg)
                    raise exceptions.TransformerConfigError(msg)

                if len(consider_true) == 0 or len(consider_false) == 0:
                    msg = "I need both `consider_true` and `consider_false` to contain at least one value."
                    logger.error(msg)
                    raise exceptions.TransformerConfigError(msg)

                common = set(consider_true) & set(consider_false)
                if common:
                   msg = f"There are values that are common to both `consider_true` and `consider_false`: {common}, this makes no sense."
                   logger.error(msg)
                   raise exceptions.TransformerConfigError(msg)

            if output_true == output_false:
                msg = "Both `output_true` and `output_false` are the same value, this makes no sense."
                logger.error(msg)
                raise exceptions.TransformerConfigError(msg)

            if output_true is None or output_false is None:
                msg = "I need both `output_true` and `output_false` to have some value."
                logger.error(msg)
                raise exceptions.TransformerConfigError(msg)

            self.output_true = output_true
            self.output_false = output_false
            self.consider_false = consider_false
            self.consider_true = consider_true

            super().__init__(raise_errors)

        def __call__(self, columns, row, i):
            for key in columns:
                value = row[key]
                if not value:
                    continue
                value = str(value)

                if value in self.consider_true:
                    boo = True
                elif value in self.consider_false:
                    boo = False
                else:
                    logger.error(f"Value `{value}` is not found in either `consider_true` (`{'`, `'.join(self.consider_true)}`) or `consider_false` (`{'`, `'.join(self.consider_false)}`). I will bypass columns {columns} at row {i}.")
                    continue

                if boo:
                    out = self.output_true
                else:
                    out = self.output_false

                logger.debug(f"Made a boolean from `{value}` to `{out}`.")
                yield out

    def __init__(self,
             properties_of,
            label_maker = None,
            branching_properties = None,
            columns=None,
            output_validator: validate.OutputValidator = None,
            multi_type_dict = None,
            raise_errors = True,
            output_true = "True",
            output_false = "False",
            consider_true = None,
            consider_false = None,
            **kwargs
        ):

        self.value_maker = self.ValueMaker(
            raise_errors=raise_errors,
            output_true    = output_true,
            output_false   = output_false,
            consider_true  = [str(i) for i in consider_true],
            consider_false = [str(i) for i in consider_false],
        )

        super().__init__(properties_of,
            self.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )


class split_translate(base.Transformer):

    def __init__(self,
        properties_of,
        label_maker = None,
        branching_properties = None,
        columns=None,
        output_validator: validate.OutputValidator = None,
        multi_type_dict = None,
        raise_errors = True,
        separator = None,
        **kwargs
    ):
        """
        FIXME doc
        """

        self.split = split(
            properties_of,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            separator = separator,
            **kwargs,
        )

        self.translate = translate(
            properties_of,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs,
        )

        super().__init__(properties_of,
            self.split.value_maker,
            label_maker,
            branching_properties,
            columns,
            output_validator,
            multi_type_dict,
            raise_errors=raise_errors,
            **kwargs
        )

    def __call__(self, row, i):

        for value in self.split.value_maker(self.split.columns, row, i):
            if base.is_not_null(value):
                logging.debug(f"VALUE {value}")
                pseudorow = {"translate_column": value}
                for val in self.translate.value_maker(["translate_column"], pseudorow, i):
                    logging.debug(f"VAL {val}")
                    value, edge_type, node_type, reverse_edge = self.create(val, row)
                    yield value, edge_type, node_type, reverse_edge



""" Base class for errors management.
"""
import logging

logger = logging.getLogger("ontoweaver")

class ErrorManager:
    def __init__(self, raise_errors = True):
        self.raise_errors = raise_errors

    def error(self, msg, section = None, index = None, exception = RuntimeError, indent = 0):
        location = ""
        if section:
            location = f" [for {section}"
            if index:
                location += f" #{index}"
            location += "]"

        err = "\t"*indent
        err += f"{exception.__name__}: "
        err += msg
        err += location

        logger.error(err)

        if self.raise_errors:
            raise exception(err)

        return err




def test_simplest():
    from . import testing_functions
    import logging
    import ontoweaver

    logging.basicConfig(level=logging.DEBUG)

    directory_name = "simplest"

    expected_nodes = [
        ('0', 'variant', {}),
        ('1', 'variant', {}),
        ('2', 'variant', {}),
        ('A', 'patient', {}),
        ('B', 'patient', {}),
        ('C', 'patient', {}),
    ]

    expected_edges = [
        ('', '0', 'A', 'patient_has_variant', {}),
        ('', '1', 'B', 'patient_has_variant', {}),
        ('', '2', 'C', 'patient_has_variant', {}),
    ]

    data_mapping = {f"tests/{directory_name}/data.csv": f"tests/{directory_name}/mapping.yaml"}

    nodes, edges = ontoweaver.extract(data_mapping, affix="none")

    logging.debug(f"NODES: {nodes}")
    logging.debug(f"EDGES: {edges}")
    bc_nodes = [n.as_tuple() for n in nodes]
    bc_edges = [e.as_tuple() for e in edges]
    fnodes, fedges = ontoweaver.fusion.reconciliate(
        bc_nodes,
        bc_edges,
        reconciliate_sep=","
    )

    logging.debug(fnodes)
    logging.debug(fedges)
    testing_functions.assert_equals(fnodes, expected_nodes)
    testing_functions.assert_equals(fedges, expected_edges)


if __name__ == "__main__":
    test_simplest()



""" Module for fusing two elements.
"""
import logging
from abc import ABCMeta, abstractmethod

from . import base
from . import merge

logger = logging.getLogger("ontoweaver")

class Fuser(merge.Merger):
    """Interface for Merger classes merging elements of a given type
    (usually a base.Element subclass).

    A subclass implementing this interface should keep trace of
    which ID was replaced by which one, in the `ID_mapping` dictionary.
    This dictionary may be used to remap the id_source and id_target of
    the related edges, processed later by another Fuser.
    """

    def __init__(self, cls):
        logger.debug(f"Instantiante {type(self).__name__} for class {cls.__name__}:")
        self.cls = cls
        self.ID_mapping = {}


class Members(Fuser):
    """A Fuser that merges base.Element objects, members by members,
    by calling the given sub-Mergers for each members.

    By definition, a merge performed with this class cannot implement
    a processing that depends on the values of other member variables.
    Each merging of member variables is independant.

    For instance, if the fusion that you want to implement decides
    how to merge types based on properties, you will need to implement
    your own subclass of Fuser.
    """

    # FIXME avoid calling methods on all 5 sub mergers each time.

    class Mergers:
        def __init__(self,
             merge_ID    = merge.string.UseKey(),
             merge_label = merge.string.UseKey(),
             merge_prop  = merge.dictry.Append(),
             merge_source= merge.string.OrderedSet(),
             merge_target= merge.string.OrderedSet()
             # merge_ID    : merge.string.StringMerger = merge.string.UseKey(),
             # merge_label : merge.string.StringMerger = merge.string.UseKey(),
             # merge_prop  : merge.dictry.DictryMerger = merge.dictry.Append(),
             # merge_source: merge.string.StringMerger = merge.string.OrderedSet(),
             # merge_target: merge.string.StringMerger = merge.string.OrderedSet()
        ):
            self.ID = merge_ID
            self.label = merge_label
            self.prop = merge_prop
            self.source = merge_source
            self.target = merge_target

    def __init__(self,
                 cls,
                 merge_ID    = merge.string.UseKey(),
                 merge_label = merge.string.UseKey(),
                 merge_prop  = merge.dictry.Append(),
                 merge_source= merge.string.OrderedSet(),
                 merge_target= merge.string.OrderedSet()
                 # merge_ID    : merge.string.StringMerger = merge.string.UseKey(),
                 # merge_label : merge.string.StringMerger = merge.string.UseKey(),
                 # merge_prop  : merge.dictry.DictryMerger = merge.dictry.Append(),
                 # merge_source: merge.string.StringMerger = merge.string.OrderedSet(),
                 # merge_target: merge.string.StringMerger = merge.string.OrderedSet()
                 ):
        """Constructor.

        Takes the class to merge and a merger for each member variable
        of either a Node or an Edge.

         Args:
             cls: either a base.Node or a base.Edge subclass
             merge_ID: the merger used to merge id(s)
             merge_label: the merger used to merge label(s)
             merge_prop: the merger used to merge property(s)
             merge_source: the merger used to merge id_source(s)
             merge_target: the merger used to merge id_target(s)
        """

        super().__init__(cls)

        self.merged = Members.Mergers(merge_ID, merge_label, merge_prop, merge_source, merge_target)
        logger.debug(f"  ID    : {type(self.merged.ID).__name__}")
        logger.debug(f"  label : {type(self.merged.label).__name__}")
        logger.debug(f"  prop  : {type(self.merged.prop).__name__}")

        self.members = {
            "id": None,
            "label": None,
            "properties": {},
        }
        if issubclass(self.cls, base.Edge):
            self.members.update({
                "id_source": None,
                "id_target": None,
            })
            logger.debug(f"  source: {type(self.merged.source).__name__}")
            logger.debug(f"  target: {type(self.merged.target).__name__}")

        self._ID_seen = set()


    def precheck(self, key, lhs, rhs):
        assert(issubclass(type(key), base.Element))
        assert(issubclass(type(lhs), base.Element))
        assert(issubclass(type(rhs), base.Element))

    def reset(self):
        self.merged.ID.reset()
        self._ID_seen = set()
        self.merged.label.reset()
        self.merged.prop.reset()
        self.merged.source.reset()
        self.merged.target.reset()

    def set(self, mergers) -> None:
        self.merged.ID.set(mergers.ID.merged)
        self.merged.label.set(mergers.label.merged)
        self.merged.prop.set(mergers.prop.merged)

        if issubclass(self.cls, base.Edge):
            self.merged.source.set(mergers.source.merged)
            self.merged.target.set(mergers.target.merged)

    def merge(self, key, lhs: base.Element, rhs: base.Element):
        assert(issubclass(type(lhs), base.Node) and issubclass(type(rhs), base.Node)
               or
               issubclass(type(lhs), base.Edge) and issubclass(type(rhs), base.Edge) )

        self._ID_seen.add(lhs.id)
        self._ID_seen.add(rhs.id)
        self.merged.ID(key, lhs.id, rhs.id)
        self.merged.label(key, lhs.label, rhs.label)
        self.merged.prop(key, lhs.properties, rhs.properties)

        if issubclass(self.cls, base.Edge):
            self.merged.source(key, lhs.id_source, rhs.id_source)
            self.merged.target(key, lhs.id_target, rhs.id_target)

        self.set(self.merged)

    def get(self) -> base.Element:
        self.members["id"] = self.merged.ID.get()
        # Save the ID mappings we've seen so far.
        for id in self._ID_seen:
            # We do not need to save self-mappings.
            if id != self.members["id"]:
                self.ID_mapping[id] = self.members["id"]

        self.members["label"] = self.merged.label.get()
        self.members["properties"] = self.merged.prop.get()

        if issubclass(self.cls, base.Edge):
            self.members["id_source"] = self.merged.source.get()
            self.members["id_target"] = self.merged.target.get()

        return self.cls(**(self.members))




""" Adapters for querying JSON data.
"""
import json
import logging
import jmespath

import pandas as pd

from typing import Optional
from itertools import chain
from enum import Enum, EnumMeta
from collections.abc import Iterable

from . import base
from . import transformer
from . import exceptions
from . import iterative
from . import validate

logger = logging.getLogger("ontoweaver")


class JSONAdapter(iterative.IterativeAdapter):
    def __init__(self,
            json_str: str,
            subject_transformer: base.Transformer,
            transformers: Iterable[base.Transformer],
            metadata: Optional[dict] = None,
            validator: Optional[validate.InputValidator] = None,
            type_affix: Optional[base.TypeAffixes] = base.TypeAffixes.suffix,
            type_affix_sep: Optional[str] = ":",
            parallel_mapping: int = 0,
            raise_errors = True,
        ):

        super().__init__(
            subject_transformer,
            transformers,
            metadata,
            validator,
            type_affix,
            type_affix_sep,
            parallel_mapping,
            raise_errors
        )

        self.json_data = json.loads(json_str)
        self.df = pd.DataFrame()

        columns = list(chain.from_iterable(
            [subject_transformer.columns]
            + [t.columns for t in transformers]
            + [k.columns for t in transformers for k in t.properties_of]
        ))

        logger.debug("JMESPath queries:")
        for c in columns:
            logger.debug(f"\t{c}")
        self.parse(columns)


    def parse(self, queries):
        assert len(queries) > 0, "No query"

        for q in queries:
            jpath = jmespath.search(q, self.json_data)
            self.df.insert(loc=0, column=q, value=list(jpath))

        logger.debug(f"JMESPath-queried DataFrame:\n{self.df}")


    def iterate(self):
        return self.df.iterrows()




""" Classes for information fusion, handling unique ID creation.
"""
from abc import ABCMeta as ABSTRACT, ABCMeta, abstractmethod

from . import base

class Serializer(metaclass=ABSTRACT):
    """Inteface for classes serializing base.Elements by producing a string
    from any part of the state of interest.

    This is used by congregate.Congretater to decide if two elements are duplicates.
    Elements having the same serialization will be considered as duplicates.
    """

    @abstractmethod
    def __call__(self, elem):
        raise NotImplementedError

class node:
    """Serializers operating on base.Node(s)."""

    class ID(Serializer):
        """Serialize a Node by using its ID."""
        def __call__(self, node):
            assert(issubclass(type(node), base.Node))
            return node.id

    class IDLabel(Serializer):
        """Serialize a Node by using its ID and label."""
        def __init__(self):
            self.id = ID()
        def __call__(self, node):
            assert(issubclass(type(node), base.Node))
            return self.id(node) + node.label

    class All(Serializer):
        """Serialize a Node by using its ID, label and properties."""
        def __init__(self):
            self.idlbl = IDLabel()
        def __call__(self, node):
            assert(issubclass(type(node), base.Node))
            return self.idlbl(node) + str(node.properties)

class edge:
    """Serializers operating on base.Edge(s)."""

    class ID(Serializer):
        """Serialize an Edge by using its ID."""
        def __call__(self, edge):
            assert(issubclass(type(edge), base.Edge))
            return edge.id

    class IDLabel(Serializer):
        """Serialize an Edge by using its ID and label."""
        def __init__(self):
            self.id = ID()
        def __call__(self, edge):
            assert(issubclass(type(edge), base.Edge))
            return self.id(edge) + edge.label

    class SourceTarget(Serializer):
        """Serialize an Edge by using its id_source and id_target."""
        def __call__(self, edge):
            assert(issubclass(type(edge), base.Edge))
            return edge.id_source + edge.id_target

    class SourceTargetLabel(Serializer):
        """Serialize an Edge by using its id_source, id_target and label."""
        def __init__(self):
            self.ST = edge.SourceTarget()
        def __call__(self, edge):
            assert(issubclass(type(edge), base.Edge))
            return self.ST(edge) + edge.label

    class All(Serializer):
        """Serialize an Edge by using its id_source, id_target, label and properties."""
        def __init__(self):
            self.idlbl = IDLabel()
            self.ST = edge.SourceTarget()
        def __call__(self, edge):
            assert(issubclass(type(edge), base.Edge))
            return self.ST(edge) + self.idlbl(edge) + str(edge.properties)


class ID(Serializer):
    """Serialize an Edge or a Node by using its ID."""
    def __init__(self):
        self.nodeid = node.ID()
        self.edgeid = edge.ID()

    def __call__(self, elem):
        if issubclass(type(elem), base.Node):
            return self.nodeid(elem)
        elif issubclass(type(elem), base.Edge):
            return self.edgeid(elem)
        else:
            assert(issubclass(type(elem), base.Node) or issubclass(type(elem), base.Edge))

class IDLabel(Serializer):
    """Serialize an Edge or a Node by using its ID and label."""
    def __init__(self):
        self.nodeidlbl = node.IDLabel()
        self.edgeidlbl = edge.IDLabel()

    def __call__(self, elem):
        if issubclass(type(elem), base.Node):
            return self.nodeidlbl(elem)
        elif issubclass(type(elem), base.Edge):
            return self.edgeidlbl(elem)
        else:
            assert(issubclass(type(elem), base.Node) or issubclass(type(elem), base.Edge))

class All(Serializer):
    """Serialize a Node or an Edge by using all their members. """
    def __init__(self):
        self.nodeall = node.All()
        self.edgeall = edge.All()

    def __call__(self, elem):
        if issubclass(type(elem), base.Node):
            return self.nodeall(elem)
        elif issubclass(type(elem), base.Edge):
            return self.edgeall(elem)
        else:
            assert(issubclass(type(elem), base.Node) or issubclass(type(elem), base.Edge))




import io
import yaml
import logging
import ontoweaver

def test_from_file():
    mapping = """
row:
   map:
       column: data[*].variant
       to_subject: variant
transformers:
    - map:
        column: data[*].patient
        to_object: patient
        via_relation: patient_has_variant
    - map:
        column: data[*].age
        to_property: age
        for_object: patient
metadata:
    - origin: somewhere
    """
    config = yaml.full_load(mapping)

    parser = ontoweaver.mapping.YamlParser(config)
    mapper = parser()

    ljs = ontoweaver.loader.LoadJSONFile()
    local_nodes,local_edges = ontoweaver.load_extract("tests/json/friends.json", config, ljs, raise_errors = False)

    for n in local_nodes:
        logging.debug(repr(n))
    for e in local_edges:
        logging.debug(repr(e))

    assert len(local_nodes) == 6
    assert len(local_edges) == 3


def test_jmespath():
    json="""
{
    "data": [
        {"variant": 0, "patient": "A", "age": 12 },
        {"variant": 1, "patient": "B", "age": 23 },
        {"variant": 2, "patient": "C", "age": 34 }
    ]
}
    """

    mapping = """
row:
   map:
       element: data[*].variant
       to_subject: variant
transformers:
    - map:
        element: data[*].patient
        to_object: patient
        via_relation: patient_has_variant
    - map:
        element: data[*].age
        to_property: age
        for_object: patient
metadata:
    - origin: somewhere
    """
    config = yaml.full_load(mapping)

    parser = ontoweaver.mapping.YamlParser(config)
    mapper = parser()

    ljs = ontoweaver.loader.LoadJSONString()
    local_nodes,local_edges = ontoweaver.load_extract(json, config, ljs, raise_errors = False)

    for n in local_nodes:
        logging.debug(repr(n))
    for e in local_edges:
        logging.debug(repr(e))

    assert len(local_nodes) == 6
    assert len(local_edges) == 3

if __name__ == "__main__":
    test_xpath()




""" How to query an XML document and iterate over the results.
"""
import logging

import pandas as pd

from lxml import etree
from typing import Optional
from itertools import chain
from enum import Enum, EnumMeta
from collections.abc import Iterable

from . import base
from . import transformer
from . import exceptions
from . import iterative
from . import validate

logger = logging.getLogger("ontoweaver")


class XMLAdapter(iterative.IterativeAdapter):
    def __init__(self,
            xml: str,
            subject_transformer: base.Transformer,
            transformers: Iterable[base.Transformer],
            metadata: Optional[dict] = None,
            validator: Optional[validate.InputValidator] = None,
            type_affix: Optional[base.TypeAffixes] = base.TypeAffixes.suffix,
            type_affix_sep: Optional[str] = ":",
            parallel_mapping: int = 0,
            raise_errors = True,
        ):

        super().__init__(
            subject_transformer,
            transformers,
            metadata,
            validator,
            type_affix,
            type_affix_sep,
            parallel_mapping,
            raise_errors
        )

        self.xml = etree.XML(xml)
        self.df = pd.DataFrame()

        columns = list(chain.from_iterable(
            [subject_transformer.columns]
            + [t.columns for t in transformers]
            + [k.columns for t in transformers for k in t.properties_of]
        ))

        logger.debug("XPath queries:")
        for c in columns:
            logger.debug(f"\t{c}")
        self.parse(columns)


    def parse(self, queries):
        assert len(queries) > 0, "No query"

        for q in queries:
            xpath = etree.ETXPath(q)
            self.df.insert(loc=0, column=q, value=[e.text for e in xpath(self.xml)])

        logger.debug(f"XPath-queried DataFrame:\n{self.df}")


    def iterate(self):
        return self.df.iterrows()



""" Parsers handling mappings.
"""
import yaml
import logging

import pandera.pandas as pa

from . import base
from . import types
from . import transformer
from . import exceptions
from . import validate
from . import make_labels

logger = logging.getLogger("ontoweaver")

class YamlParser(base.MappingParser):
    """
    Parse a table extraction configuration and return the three objects needed to configure an Adapter.

    The config is a dictionary containing only strings, as converted from the following YAML description:

    .. code-block:: yaml

        row:
           map:
              columns:
                - <MY_COLUMN_NAME>
              to_subject: <MY_SUBJECT_TYPE>
        transformers:
            - map:
                columns:
                    - <MY_COLUMN_NAME>
                to_object: <MY_OBJECT_TYPE>
                via_relation: <MY_RELATION_TYPE>
            - map:
                columns:
                    - <MY_OTHER_COLUMN>
                to_property:
                    - <MY_PROPERTY>
                for_objects:
                    - <MY_OBJECT_TYPE>

    This maps the table row to a MY_SUBJECT_TYPE node type, adding an edge of type MY_RELATION_TYPE,
    between the MY_SUBJECT_TYPE node and another MY_OBJECT_TYPE node. The data in MY_OTHER_COLUMN is mapped
    to the MY_PROPERTY property of the MY_OBJECT_TYPE node. Note that `to_properties` may effectively map to
    an edge type or several types.

    In order to allow the user to write mappings configurations using their preferred vocabulary, the following
    keywords are interchangeable:

        - subject = row = entry = line,
        - columns = fields,
        - to_target = to_object = to_node
        - via_edge = via_relation = via_predicate.

    :param dict config: A configuration dictionary.
    :param module: The module in which to insert the types declared by the configuration.
    :return tuple: subject_transformer, transformers, metadata as needed by the Adapter.
    """
    def __init__(self, config: dict, module = None, validate_output = False, raise_errors = True):
        """
        Initialize the YamlParser.

        Args:
            config (dict): The configuration dictionary.
            validate_output (bool): Whether to validate the output of the transformers. Defaults to False.
            module: The module in which to insert the types declared by the configuration.
        """
        super().__init__(module, raise_errors = raise_errors)
        self.config = config
        self.validate_output = validate_output
        if not module:
            self.module = types
        else:
            self.module = module

        if not self.validate_output:
            logger.info(
                f"Transformer output validation will be skipped. This could result in some empty or `nan` nodes in your knowledge graph."
                f" To enable output validation set `validate_output` to `True`.")

        logger.debug(f"Classes will be created in module '{self.module}'")


    def make_transformer(self, transformer_type, multi_type_dictionary = None, branching_properties = None,
                               properties=None, columns=None, output_validator=None, label_maker = None, **kwargs):
        """
        LabelMaker a transformer class and instance with the given parameters.

        Args:
            multi_type_dictionary: Dictionary of regex rules and corresponding types in case of cell value match.
            transformer_type: The class of the transformer.
            properties: The properties of the transformer.
            columns: The columns to be processed by the transformer.
            output_validator: validate.OutputValidator instance for transformer output validation.
            **kwargs: Additional keyword arguments.

        Returns:
            The created transformer instance.

        Raises:
            TypeError: If the transformer type is not an existing transformer.
        """
        # If `transformer` module has this specific type.
        if hasattr(transformer, transformer_type):
            class_t = getattr(transformer, transformer_type)
            kwargs.setdefault("subclass", class_t)
            if not issubclass(class_t, base.Transformer):
                self.error(f"Object `{transformer_type}` is not an existing transformer.", exception = exceptions.DeclarationError)
            logger.debug(f"\t\tDeclare Transformer class '{transformer_type}'.")
            return class_t(
                properties_of = properties,
                label_maker = label_maker,
                branching_properties = branching_properties,
                columns = columns,
                output_validator = output_validator,
                multi_type_dict = multi_type_dictionary,
                raise_errors = self.raise_errors,
                **kwargs
            )
        else:
            # logger.debug(dir(generators))
            self.error(f"Cannot find a transformer class with name `{transformer_type}`.", exception = exceptions.DeclarationError)


    def _get_input_validation_rules(self,):
        """
        Extract input data validation schema from yaml file and instantiate a Pandera DataFrameSchema object and validator.
        """
        k_validate = ["validate"]
        validation_rules = self.get(k_validate)
        yaml_validation_rules = yaml.dump(validation_rules, default_flow_style=False)
        validator = None

        try:
            validation_schema = pa.DataFrameSchema.from_yaml(yaml_validation_rules)
            validator = validate.InputValidator(validation_schema, raise_errors=self.raise_errors)
        except Exception as e:
            self.error(f"Failed to parse the input validation schema: {e}", exception=exceptions.ConfigError)

        return validator


    # ====================================================================
    # Getter functions for retrieval of dictionary items from YAML config
    # ====================================================================

    def get_not(self, keys, pconfig=None):
        """
        Get the first dictionary (key, item) not matching any of the passed keys.

        Args:
            keys: The keys to exclude.
            pconfig: The configuration dictionary to search in (default is self.config).

        Returns:
            dict: The first dictionary not matching any of the passed keys.
        """
        res = {}
        if not pconfig:
            pconfig = self.config
        for k in pconfig:
            if k not in keys:
                res[k] = pconfig[k]
        return res

    def get(self, keys, pconfig=None):
        """
        Get a dictionary item matching any of the passed keys.

        Args:
            keys: The keys to search for.
            pconfig: The configuration dictionary to search in (default is self.config).

        Returns:
            The first item matching any of the passed keys, or None if no match is found.
        """
        if not pconfig:
            pconfig = self.config
            assert pconfig, "I must have a configuration dictionary."
        for k in keys:
            if k in pconfig:
                return pconfig[k]
        return None


    # ================================================================
    # Helper Functions for both Target, Subject and Property parsing
    # ================================================================

    def _extract_metadata(self, k_metadata_column, metadata_list, metadata, types, columns):
        """
        Extract metadata and update the metadata dictionary.

        Args:
            k_metadata_column (list): List of keys to be used for adding source column names.
            metadata_list (list): List of metadata items to be added.
            metadata (dict): The metadata dictionary to be updated.
            types (str): The type of the node or edge.
            columns (list): List of columns to be added to the metadata.

        Returns:
            dict: The updated metadata dictionary.
        """
        if metadata_list and types:
            if not isinstance(types, set):
                t = types
                metadata.setdefault(t, {})
                for item in metadata_list:
                    metadata[t].update(item)
                for key in k_metadata_column:
                    if key in metadata[t]:
                        # Use the value of k_metadata_column as the key.
                        key_name = metadata[t][key]
                        # Remove the k_metadata_column key from the metadata dictionary.
                        if key_name in metadata[t]:
                            msg = f"The key you used for adding source column names: `{key_name}` to node: `{t}` already exists in the metadata dictionary."
                            # FIXME is it an error or a warning?
                            # self.error(msg)
                            logger.warning(msg)
                        del metadata[t][key]
                        if columns:
                            # TODO make the separator a parameter.
                            metadata[t][key_name] = ", ".join(columns)
            else:
                for t in types:
                    metadata.setdefault(t, {})
                    for item in metadata_list:
                        metadata[t].update(item)
                    for key in k_metadata_column:
                        if key in metadata[t]:
                            # Use the value of k_metadata_column as the key.
                            key_name = metadata[t][key]
                            # Remove the k_metadata_column key from the metadata dictionary.
                            if key_name in metadata[t]:
                                msg = f"The key you used for adding source column names: `{key_name}` to node: `{t}` already exists in the metadata dictionary."
                                # FIXME is it an error or a warning?
                                # self.error(msg)
                                logger.warning(msg)
                            del metadata[t][key]
                            if columns:
                                # TODO make the separator a parameter.
                                metadata[t][key_name] = ", ".join(columns)

            return metadata
        else:
            return None

    def _make_output_validator(self, output_validation_rules = None):
        """
        LabelMaker a validator for the output of a transformer.

        Args:
            output_validation_rules: The output validation rules for the transformer extracted from yaml file.

        Returns:
            validate.OutputValidator: The created validator.
        """

        if self.validate_output:
            if output_validation_rules:
                output_validator = validate.OutputValidator(raise_errors=self.raise_errors)
                # Adjust the formatting for output validation rules to match the expected format. This is so the
                # user would not have to type `columns` and `cell_value` in the configuration file each time.
                dict_output_validation_rules = {"columns": {"cell_value": output_validation_rules}}
                yaml_output_validation_rules = yaml.dump(dict_output_validation_rules, default_flow_style=False)
                output_validator.update_rules(pa.DataFrameSchema.from_yaml(yaml_output_validation_rules))
            else:
                output_validator = validate.SimpleOutputValidator(raise_errors=self.raise_errors)
        else:
            output_validator = validate.SkipValidator(raise_errors=self.raise_errors)

        return output_validator


    def _extract_final_type_class(self, final_type, possible_types, metadata, metadata_list, columns, properties_of):
        """
        Extract metadata and class for the final type and update the metadata dictionary.
        """

        if final_type:
            final_type_class = self.make_node_class(
                final_type,
                properties_of.get(final_type, {})
            )
            possible_types.add(final_type_class.__name__)
            extracted_s_final_type_metadata = self._extract_metadata(
                base.MappingParser.k_metadata_column,
                metadata_list, metadata,
                final_type,
                columns )

            if extracted_s_final_type_metadata:
                metadata.update(extracted_s_final_type_metadata)

            return final_type_class

        return None


    def _make_branching_dict(self, subject: bool, match_parser, properties_of, metadata_list, metadata, columns, final_type_class,
                             multi_type_dictionary, possible_node_types, possible_edge_types):
        """
        Helper function to parse the `match` clause of the YAML configuration file for subject and target transformers.
        """

        k_extract  = base.MappingParser.k_subject if subject else base.MappingParser.k_target

        for entry in match_parser:
            for k, v in entry.items():
                if isinstance(v, dict):
                    key = k
                    multi_type_dictionary[key] = {k1: v1 for k1, v1 in v.items()}
                    alt_type = self.get(k_extract, v)
                    alt_type_class = self.make_node_class(alt_type, properties_of.get(alt_type, {}))

                    possible_node_types.add(alt_type)

                    alt_final_type = self.get(base.MappingParser.k_final_type, v)

                    alt_final_type_class = self._extract_final_type_class(
                        alt_final_type,
                        possible_node_types,
                        metadata,
                        metadata_list,
                        columns,
                        properties_of
                    )

                    if not subject:
                        alt_edge = self.get(base.MappingParser.k_edge, v)
                        alt_edge_class = self.make_edge_class(
                            alt_edge,
                            None,
                            alt_type_class,
                            properties_of.get(alt_edge, {})
                        )

                        possible_edge_types.add(alt_edge)

                        extracted_alt_edge_metadata = self._extract_metadata(
                            base.MappingParser.k_metadata_column,
                            metadata_list, metadata,
                            alt_edge,
                            None
                        )

                        if extracted_alt_edge_metadata:
                            metadata.update(extracted_alt_edge_metadata)

                        # Extract reverse edge, if specified in config.
                        alt_reverse_edge = self.get(base.MappingParser.k_reverse_edge, v)
                        if alt_reverse_edge:
                            alt_reverse_edge_class = self.make_edge_class(
                                alt_reverse_edge,
                                None,
                                alt_type_class,
                                properties_of.get(alt_reverse_edge, {})
                            )

                            possible_edge_types.add(alt_reverse_edge)
                            extracted_alt_reverse_edge_metadata = self._extract_metadata(
                                base.MappingParser.k_metadata_column,
                                metadata_list,
                                metadata,
                                alt_reverse_edge,
                                None
                            )

                            if extracted_alt_reverse_edge_metadata:
                                metadata.update(extracted_alt_reverse_edge_metadata)

                        #TODO: Create new function or add this to make edge class?

                    extracted_alt_type_metadata = self._extract_metadata(
                        base.MappingParser.k_metadata_column,
                        metadata_list,
                        metadata,
                        alt_type,
                        columns
                    )
                    if extracted_alt_type_metadata:
                        metadata.update(extracted_alt_type_metadata)

                    if extracted_alt_type_metadata:
                        metadata.update(extracted_alt_type_metadata)

                    assert alt_type_class
                    multi_type_dictionary[key] = {
                        'to_object': alt_type_class,
                        # Via relation is always None for subject, since there is never an edge declared for the subject type.
                        'via_relation': None if subject is True else alt_edge_class,
                        # We first try and declare the final type passed to the whole class of the
                        # transformer, and if it is not defined, we use the final type of the
                        # alternative node type, under the `match` clause.
                        'final_type': final_type_class if final_type_class else alt_final_type_class,
                        # None if subject because subject does not come with edge, and None if alt reverse edge class is not
                        # declared in mapping file.
                        'reverse_relation': None if subject is True else alt_reverse_edge_class if alt_reverse_edge else None,
                    }


    # ============================================================
    # Property parsing
    # ============================================================

    def parse_properties(self, properties_of, possible_subject_types, transformers_list):
        """
        Parse the properties of the transformers defined in the YAML mapping, and update the properties_of dictionary.
        """
        logger.debug(f"Parse properties...")

        for n_transformer, transformer_types in enumerate(transformers_list):
            if not hasattr(transformer_types, "items"):
                transformer_types = {transformer_types: []}

            for transformer_type, field_dict in transformer_types.items():
                if not field_dict:
                    logger.warning(f"There is no field for the {n_transformer}th transformer: '{transformer_type}',"
                               f" did you forget an indentation?"
                               f" If it is a User defined trasnformer, please disregard this warning. ")

                    continue

                if any(field in field_dict for field in base.MappingParser.k_properties):
                    object_types = self.get(base.MappingParser.k_prop_to_object, pconfig=field_dict)
                    property_names = self.get(base.MappingParser.k_properties, pconfig=field_dict)
                    if not isinstance(property_names, list):
                        logger.debug(f"\tDeclared singular property")
                        assert isinstance(property_names, str)
                        property_names = [property_names]
                    if not object_types:  # FIXME: Creates errors with branching subject types and subject final type features.
                        logger.info(
                            f"No `for_objects` defined for properties {property_names}, I will attach those properties to the row subject(s) `{possible_subject_types}`")
                        if isinstance(possible_subject_types, set) and len(possible_subject_types) == 1:
                            object_types = list(possible_subject_types)[0]
                        if isinstance(possible_subject_types, set) and len(possible_subject_types) > 1:
                            object_types = list(possible_subject_types)
                    if not isinstance(object_types, list):
                        logger.debug(f"\tDeclared singular for_object: `{object_types}`")
                        assert (isinstance(object_types, str))
                        object_types = [object_types]

                    column_names = self.get(base.MappingParser.k_columns, pconfig=field_dict)
                    if column_names != None and not isinstance(column_names, list):
                        logger.debug(f"\tDeclared singular column `{column_names}`")
                        assert (isinstance(column_names, str))
                        column_names = [column_names]
                    gen_data = self.get_not(base.MappingParser.k_target + base.MappingParser.k_edge + base.MappingParser.k_columns, pconfig=field_dict)

                    # Parse the validation rules for the output of the property transformer.
                    p_output_validation_rules = self.get(base.MappingParser.k_validate_output, pconfig=field_dict)
                    p_output_validator = self._make_output_validator(p_output_validation_rules)

                    prop_transformer = self.make_transformer(
                        transformer_type, columns=column_names,
                        output_validator=p_output_validator,
                        label_maker=make_labels.SimpleLabelMaker(
                            raise_errors=self.raise_errors),
                        **gen_data)

                    for object_type in object_types:
                        properties_of.setdefault(object_type, {})
                        for property_name in property_names:
                            properties_of[object_type].setdefault(prop_transformer, property_name)
                        logger.debug(f"\t\tDeclared property mapping for `{object_type}`: {properties_of[object_type]}")

        return properties_of

    # ============================================================
    # Subject parsing
    # ============================================================

    def parse_subject(self, properties_of, transformers_list, metadata_list, metadata):
        """
        Parse the subject transformer and its properties from the YAML mapping.
        """

        logger.debug("Declare subject type...")
        subject_transformer_dict = self.get(base.MappingParser.k_row)
        if not subject_transformer_dict:
            msg = f"There is no `{'`, `'.join(base.MappingParser.k_row)}` key in your mapping."
            logging.error(msg)
            raise RuntimeError(msg)

        subject_transformer_class = list(subject_transformer_dict.keys())[0]
        subject_kwargs = self.get_not(base.MappingParser.k_subject_type + base.MappingParser.k_columns, subject_transformer_dict[
            subject_transformer_class])  # FIXME shows redundant information filter out the keys that are not needed.
        subject_columns = self.get(
            base.MappingParser.k_columns,
            subject_transformer_dict[subject_transformer_class] )
        subject_final_type = self.get(
            base.MappingParser.k_final_type,
            subject_transformer_dict[subject_transformer_class] )
        if subject_columns != None and not isinstance(subject_columns, list):
            logger.debug(f"\tDeclared singular subject’s column `{subject_columns}`")
            assert (isinstance(subject_columns, str))
            subject_columns = [subject_columns]

        # Parse the validation rules for the output of the subject transformer.
        subject_output_validation_rules = self.get(
            base.MappingParser.k_validate_output,
            subject_transformer_dict[subject_transformer_class] )

        subject_output_validator = self._make_output_validator(subject_output_validation_rules)

        subject_multi_type_dict = {}
        # TODO: assert subject_transformer_dict contains only ONE transformer.

        subject_transformer_params = subject_transformer_dict[subject_transformer_class]

        possible_subject_types = set()
        subject_branching = False

        # NOTE: this alters possible_subject_types and metadata
        s_final_type_class = self._extract_final_type_class(
            subject_final_type, possible_subject_types, metadata,
            metadata_list, subject_columns, properties_of)

        if subject_transformer_params:
            logger.debug("Subject section has parameters.")

            if "match" in subject_transformer_params:
                logger.debug("\tThere is a `match` in the subject section.")

                subject_branching = True

                self._make_branching_dict(
                    subject = True,
                    match_parser = subject_transformer_params["match"],
                    properties_of = properties_of,
                    metadata_list = metadata_list, metadata = metadata,
                    columns = subject_columns,
                    final_type_class = s_final_type_class,
                    multi_type_dictionary = subject_multi_type_dict,
                    possible_node_types = possible_subject_types,
                    possible_edge_types = set()
                )

                source_t = None # Subject_type declared None because the subject transformer is a branching transformer.
                subject_type = None
                logger.debug("Parse subject transformer...")

                if any(key in subject_kwargs for key in base.MappingParser.k_match_type_from):
                    s_label_maker = make_labels.MultiTypeOnColumnLabelMaker(
                        raise_errors=self.raise_errors,
                        match_type_from_column=self.get(base.MappingParser.k_match_type_from, subject_kwargs)
                    )
                else:
                    s_label_maker = make_labels.MultiTypeLabelMaker(raise_errors=self.raise_errors)

            # "None" key is used to return any type of string, in case no branching is needed.
            else:
                logger.debug("\tThere is NO `match` in the subject section.")

                subject_type = self.get(
                    base.MappingParser.k_subject_type,
                    subject_transformer_dict[subject_transformer_class]
                )
                source_t = self.make_node_class(
                    subject_type,
                    properties_of.get(subject_type, {})
                )
                assert source_t

                subject_multi_type_dict = {'None': {
                    'to_object': source_t,
                    'via_relation': None,
                    'final_type': s_final_type_class,
                    'reverse_relation': None
                }}

                possible_subject_types.add(source_t.__name__)

                s_label_maker = make_labels.SimpleLabelMaker( raise_errors = self.raise_errors )

            # Append properties to subject type(s). In case other types are not declared.
            properties_of = self.parse_properties(properties_of, possible_subject_types, transformers_list)

            subject_transformer = self.make_transformer(
                transformer_type = subject_transformer_class,
                multi_type_dictionary = subject_multi_type_dict,
                branching_properties = properties_of if subject_branching else None,
                properties = properties_of.get(
                    subject_type,
                    {}
                ) if not subject_branching else None,
                columns = subject_columns,
                output_validator = subject_output_validator,
                label_maker = s_label_maker,
                **subject_kwargs
            )

            logger.debug(f"\tDeclared subject transformer: {subject_transformer}")

            logger.debug(
                f"\tDeclare subject of possible types: `{possible_subject_types}`, subject transformer: `{subject_transformer_class}`, "
                f"subject kwargs: `{subject_kwargs}`, subject columns: '{subject_columns}'")

        else:
            logger.debug("\tSubject section does not have parameters.")

            # Still parse properties even if no subject transformer parameters are declared.
            properties_of = self.parse_properties(properties_of, possible_subject_types, transformers_list)

            logger.warning(f"No keywords declared for subject transformer `{subject_transformer_class}`"
                       f" Creating transformer with no parameters.")

            subject_transformer = self.make_transformer(
                transformer_type = subject_transformer_class,
                 branching_properties=properties_of
             )

            # Declare source type as None because no parameters were declared for the subject transformer.
            source_t = None

        extracted_metadata = self._extract_metadata(
            base.MappingParser.k_metadata_column,
            metadata_list,
            metadata,
            possible_subject_types,
            subject_columns
        )
        if extracted_metadata:
            metadata.update(extracted_metadata)

        return possible_subject_types, subject_transformer, source_t, subject_columns

    # ============================================================
    # Target Transformer Parsing and Helper Functions
    # ============================================================

    def _make_target_label_maker(self, target, edge, gen_data, columns, transformer_index, multi_type_dictionary):

        label_maker = None

        if (target and not edge) or (edge and not target):
            self.error(f"Cannot declare the mapping  `{columns}` => `{edge}` (target: `{target}`), "
                       f"missing either a `to_object` or a `via_relation` in the mapping?.", "transformers", transformer_index,
                       indent=2, exception=exceptions.MissingDataError)

        elif multi_type_dictionary and any(key in gen_data for key in base.MappingParser.k_match_type_from) and not target and not edge:
            label_maker = make_labels.MultiTypeOnColumnLabelMaker(
                raise_errors=self.raise_errors,
                match_type_from_column=self.get(base.MappingParser.k_match_type_from, gen_data)
            )

        elif multi_type_dictionary and not target and not edge and not any(key in gen_data for key in base.MappingParser.k_match_type_from):
            label_maker = make_labels.MultiTypeLabelMaker(raise_errors=self.raise_errors)
        else:
            label_maker = make_labels.SimpleLabelMaker(raise_errors = self.raise_errors)

        return label_maker

    def _check_target_sanity(self, transformer_keyword_dict, transformer_type, transformer_index, transformers, properties_of):
        """
        Preforms sanity checks on target transformer before assigning the target, edge and subject variables, and connected
        columns.
        """

        if not transformer_keyword_dict:
            logger.warning(f"No keywords declared for target transformer `{transformer_type}` at index"
                       f" `{transformer_index}`. Creating transformer with no parameters.")
            target_transformer = self.make_transformer(transformer_type=transformer_type,
                                                             branching_properties=properties_of)
            transformers.append(target_transformer)

        elif any(field in transformer_keyword_dict for field in base.MappingParser.k_properties):
            if any(field in transformer_keyword_dict for field in base.MappingParser.k_target):
                prop = self.get(base.MappingParser.k_properties, transformer_keyword_dict)
                target = self.get(base.MappingParser.k_target, transformer_keyword_dict)
                col = self.get(base.MappingParser.k_columns, transformer_keyword_dict)
                self.error(f"ERROR in transformer '{transformer_type}'," \
                           f" at index `{transformer_index}`: one cannot " \
                           f" declare a mapping to both properties '{prop}'" \
                           f" and object type '{target}'." \
                           f" Maybe you used `to_object` instead of `for_object`" \
                            " to indicate to which object to attach the property?",
                           "transformers",
                           transformer_index,
                           exception=exceptions.CardinalityError
                       )

        elif not isinstance(transformer_keyword_dict, dict):
            self.error(str(transformer_keyword_dict) + " is not a dictionary",
                       exception=exceptions.ParsingDeclarationsError)

        else:
            columns = self.get(base.MappingParser.k_columns, pconfig=transformer_keyword_dict)
            if not isinstance(columns, list):
                logger.debug("\tDeclared singular column")
                # The rowIndex transformer is a special case, where the column does not need to be defined in the mapping.
                # FIXME: In next refactoring do not assert `rowIndex` transformer name, in order to have a generic implementation. (ref: https://github.com/oncodash/ontoweaver/pull/153)
                if not columns:
                    columns = []
                elif transformer_type != "rowIndex":
                    assert (isinstance(columns, str))
                    columns = [columns]

            target = self.get(base.MappingParser.k_target, pconfig=transformer_keyword_dict)
            if isinstance(target, list):
                self.error(
                    f"You cannot declare multiple objects in transformers. For transformer `{transformer_type}`.",
                    section="transformers", index=transformer_index, indent=1, exception=exceptions.CardinalityError)

            subject = self.get(base.MappingParser.k_subject, pconfig=transformer_keyword_dict)
            if isinstance(subject, list):
                self.error(
                    f"You cannot declare multiple subjects in transformers. For transformer `{transformer_type}`.",
                    section="transformers", index=transformer_index, indent=1, exception=exceptions.CardinalityError)

            edge = self.get(base.MappingParser.k_edge, pconfig=transformer_keyword_dict)
            if isinstance(edge, list):
                self.error(
                    f"You cannot declare multiple relations in transformers. For transformer `{transformer_type}`.",
                    section="transformers", index=transformer_index, indent=1, exception=exceptions.CardinalityError)

            reverse_relation = self.get(base.MappingParser.k_reverse_edge, pconfig=transformer_keyword_dict)
            if isinstance(reverse_relation, list):
                self.error(
                    f"You cannot declare multiple reverse relations in transformers. For transformer `{transformer_type}`.",
                    section="transformers", index=transformer_index, indent=1, exception=exceptions.CardinalityError)

            return columns, target, edge, subject, reverse_relation


    def _make_target_classes(self, target, properties_of, edge, source_t, final_type_class, reverse_relation, possible_target_types, possible_edge_types, multi_type_dictionary):
        """
        Helper function to create the target and edge classes for a target transformer, and store them in the multi_type_dictionary.
        """

        logger.debug(f"\tDeclare node target `{target}`...")
        target_t = self.make_node_class(target, properties_of.get(target, {}))
        possible_target_types.add(target)
        logger.debug(f"\t\tDeclared target`{target}`: {target_t.__name__}")

        edge_t = self.make_edge_class(edge, source_t, target_t, properties_of.get(edge, {}))
        logger.debug(f"\tDeclare edge `{edge}: {edge_t.__name__}")
        possible_edge_types.add(edge_t.__name__)

        if reverse_relation is not None:
            reverse_relation_t = self.make_edge_class(reverse_relation, target_t, source_t, properties_of.get(reverse_relation, {}))
            logger.debug(f"\tDeclare reverse relation `{reverse_relation}: {reverse_relation_t.__name__}")
            possible_edge_types.add(reverse_relation_t)



        # "None" key is used to return any type of string, in case no branching is needed.
        multi_type_dictionary['None'] = {
            'to_object': target_t,
            'via_relation': edge_t,
            'final_type': final_type_class,
            'reverse_relation': reverse_relation_t if reverse_relation is not None else None,
        }

        # FIXME: Edge may not be needed as return here.
        return edge_t, target_t

    # ============================================================
    # Target parsing
    # ============================================================

    def parse_targets(self, transformers_list, properties_of, source_t, metadata_list, metadata):
        """
        Parse the target transformers and their properties from the YAML mapping.
        """

        transformers = []
        possible_target_types = set()
        possible_edge_types = set()

        logger.debug("Declare types...")
        # Iterate through the list of target transformers and extract the information needed to create the
        # corresponding classes.
        for transformer_index, target_transformer_yaml_dict in enumerate(transformers_list):
            if not hasattr(target_transformer_yaml_dict, "items"):
                target_transformer_yaml_dict = {target_transformer_yaml_dict: []}

            for transformer_type, transformer_keyword_dict in target_transformer_yaml_dict.items():

                target_branching = False
                elements = self._check_target_sanity(transformer_keyword_dict, transformer_type, transformer_index, transformers, properties_of)
                if elements is None:
                    continue
                # Transformer passed sanity check, unpack the returned elements.
                else:
                    # FIXME double-check why subject is unused here.
                    columns, target, edge, subject, reverse_relation = elements

                gen_data = self.get_not(base.MappingParser.k_target + base.MappingParser.k_edge + base.MappingParser.k_columns + base.MappingParser.k_final_type + base.MappingParser.k_reverse_edge, pconfig=transformer_keyword_dict)

                # Extract the final type if defined in the mapping.
                final_type = self.get(base.MappingParser.k_final_type, pconfig=transformer_keyword_dict)
                final_type_class = self._extract_final_type_class(final_type, possible_target_types, metadata,
                                                                  metadata_list, columns, properties_of)

                # Harmonize the use of the `from_subject` and `from_source` synonyms in the configuration, because
                # from_subject` is used in the transformer class to refer to the source node type.
                if 'from_source' in gen_data:
                    gen_data['from_subject'] = gen_data['from_source']
                    del gen_data['from_source']

                multi_type_dictionary = {}

                # The target transformer is a simple transformer
                # if it does not have a `match` clause.
                # We create a simple multi_type_dictionary,
                # with a "None" key, to indicate that no branching is needed.
                if target and edge:
                    # FIXME double-check why edge_t is unused here.
                    edge_t, target_t = self._make_target_classes(target, properties_of, edge, source_t, final_type_class, reverse_relation, possible_target_types, possible_edge_types, multi_type_dictionary)
                    # Parse the validation rules for the output of the transformer.
                    # Each transformer gets its own
                    # instance of the OutputValidator with (at least) the
                    # default output validation rules.
                    output_validation_rules = self.get(base.MappingParser.k_validate_output, pconfig=transformer_keyword_dict)
                    # output_validator = self._make_output_validator(output_validation_rules)
                    logger.debug(f"\tDeclare transformer `{transformer_type}`...")

                # The target transformer is a branching transformer
                # if it has a `match` clause.
                # We create a branching dictionary.
                # The keys of the dictionary are the regex patterns to be matched
                # against the extracted value of the column.
                if "match" in gen_data:

                    target_branching = True
                    self._make_branching_dict(
                        subject = False,
                        match_parser = gen_data["match"],
                        properties_of = properties_of,
                        metadata_list = metadata_list,
                        metadata = metadata,
                        columns = columns,
                        final_type_class = final_type_class,
                        multi_type_dictionary = multi_type_dictionary,
                        possible_node_types = possible_target_types,
                        possible_edge_types = possible_edge_types
                    )

                # Parse the validation rules for the output of the transformer. Each transformer gets its own
                # instance of the OutputValidator with (at least) the default output validation rules.
                output_validation_rules = self.get(base.MappingParser.k_validate_output, pconfig=transformer_keyword_dict)
                output_validator = self._make_output_validator(output_validation_rules)

                label_maker = self._make_target_label_maker(target, edge, gen_data, columns, transformer_index, multi_type_dictionary)
                target_transformer = self.make_transformer(
                    transformer_type = transformer_type,
                    multi_type_dictionary = multi_type_dictionary,
                    branching_properties = properties_of if target_branching else None,
                    properties = properties_of.get(
                        target_t.__name__,
                        {}) if not target_branching else None,
                    columns = columns,
                    output_validator = output_validator,
                    label_maker = label_maker,
                    **gen_data
                )
                transformers.append(target_transformer)
                # Declare the metadata for the target and edge types.

                extracted_metadata = self._extract_metadata(base.MappingParser.k_metadata_column, metadata_list, metadata, target, columns)
                if extracted_metadata:
                    metadata.update(extracted_metadata)
                if edge:
                    extracted_metadata = self._extract_metadata(base.MappingParser.k_metadata_column, metadata_list, metadata, edge, None)
                    if extracted_metadata:
                        metadata.update(extracted_metadata)

        return transformers, possible_target_types, possible_edge_types, metadata


    def __call__(self):
        """
        Parse the configuration and return the subject transformer and transformers.

        Returns:
            tuple: The subject transformer and a list of transformers.
        """
        logger.debug("Parse mapping...")

        properties_of = {}
        metadata = {}

        #TODO create make node class for nested final_type instantiation

        # Extract transformer list and metadata list from the config.
        transformers_list = self.get(base.MappingParser.k_transformer)
        if not transformers_list:
            self.error("I cannot find the `transformers' section or it is empty," \
                        " check syntax for a typo (forgotten `s'?)" \
                        " or declare at least one transformer.`",
            exception = exceptions.ParsingDeclarationsError)

        metadata_list = self.get(base.MappingParser.k_metadata)

        # Parse subject type, metadata for subject, and properties for both subject and target types (parse_subject calls parse_properties).
        possible_subject_types, subject_transformer, source_t, subject_columns = self.parse_subject(properties_of, transformers_list, metadata_list, metadata)

        # Parse the target types and target metadata.
        transformers, possible_target_types, possible_edge_types, metadata = self.parse_targets(transformers_list, properties_of, source_t, metadata_list, metadata)

        validator = self._get_input_validation_rules()

        if not validator or isinstance(validator, validate.SkipValidator):
            logger.warning(f"Input validation will be skipped.")

        skipped_columns = []
        for t in transformers:
            if isinstance(t.output_validator, validate.SkipValidator):
                skipped_columns += t.columns

        if skipped_columns:
            logger.warning(f"Skip output validation for columns: `{'`, `'.join(str(i) for i in skipped_columns)}`." \
                           " This could result in some empty or `nan` nodes." \
                           " To enable output validation set `validate_output` to `True`.")

        if source_t:
            logger.debug(f"source class: {source_t}")
        elif possible_edge_types:
            logger.debug(f"possible_source_types: {possible_subject_types}")

        logger.debug("properties_of:")
        for kind in properties_of:
            for k in properties_of[kind]:
                logger.debug(f"\t{kind}: {k} {properties_of[kind][k]}>")

        logger.debug("transformers:")
        for t in transformers:
            logger.debug(f"\t{t}")

        if len(metadata) > 0:
            logger.debug("metadata:")
            for k in metadata:
                logger.debug(f"\t{metadata[k]}")
        else:
            logging.debug("No metadata")

        return subject_transformer, transformers, metadata, validator





import re
import os
import sys
import logging
import pathlib
import subprocess

def extract_code_from_doc(docfile, demofile, prefix):
    with open(docfile) as fd:
        text = fd.readlines()

    section = []
    i = 0
    while i < len(text):
        line = text[i]
        # print(line, end="")
        if not re.match(r"^\s{4}:caption:\s" + prefix + demofile + r"$", line):
            i += 1
            continue
        else:
            logging.debug(f"\tFound: {line.strip()}")
            line = "    "
            i += 1
            while line != "":
                has_code = re.search(r"^\s{4}(.+)$", line)
                if has_code:
                    code = has_code.group(1)
                    section.append(code)
                    line = text[i]
                    i += 1
                elif re.match(r"\s{4}$", line):
                    line = text[i]
                    i += 1
                    continue
                else:
                    break
            break

    return "\n".join(section)


def doc_to_file(docfile, demofile, dir, prefix):
    outfile = pathlib.Path(dir) / pathlib.Path(demofile)
    logging.debug(f"Parsing {docfile} for {prefix}{demofile}")
    content = extract_code_from_doc(docfile, demofile, prefix)
    assert content, "Unable to extract a file content."
    logging.debug(f"\tWriting extracted code to: {outfile}")
    with open(outfile, 'w') as fd:
        fd.write(content)


def test_full_fledged_demo():
    tutorials = "docs/sections/tutorials.rst"
    dir = "tests/full_fledge_demo"
    prefix = "demo_"
    logging.warning("This test extracts the code (mapping, ontology, etc.) from the documentation")
    logging.warning("If you want to fix some bug, DO NOT FIX THE FILES in test/, but instead, fix the documentation in `tutorial.rst`.")
    doc_to_file(tutorials, "data.csv"    , dir, prefix)
    doc_to_file(tutorials, "ontology.ttl", dir, prefix)
    doc_to_file(tutorials, "config.yaml" , dir, prefix)
    doc_to_file(tutorials, "mapping.yaml", dir, prefix)
    doc_to_file(tutorials, "schema.yaml" , dir, prefix)

    xs = f"{dir}/extended_schema.yaml"
    if os.path.isfile(xs):
        os.remove(xs)

    cmd = f"ontoweave --biocypher-config {dir}/config.yaml --auto-schema {xs} --biocypher-schema {dir}/schema.yaml {dir}/data.csv:{dir}/mapping.yaml --auto-schema-overwrite --debug"

    logging.debug(f"Run: {cmd}")
    subprocess.run(cmd.split(), capture_output=True, check=True)


def test_simplest_example():
    tutorials = "docs/sections/tutorials.rst"
    dir = "tests/simplest_example"
    prefix = "simple_"
    doc_to_file(tutorials, "data.csv"    , dir, prefix)
    doc_to_file(tutorials, "ontology.ttl", dir, prefix)
    doc_to_file(tutorials, "config.yaml" , dir, prefix)
    doc_to_file(tutorials, "mapping.yaml", dir, prefix)

    xs = f"{dir}/extended_schema.yaml"
    if os.path.isfile(xs):
        os.remove(xs)

    cmd = f"ontoweave --biocypher-config {dir}/config.yaml --auto-schema {xs} {dir}/data.csv:{dir}/mapping.yaml --auto-schema-overwrite --debug"

    logging.debug(f"Run: {cmd}")
    subprocess.run(cmd.split(), capture_output=True, check=True)


if __name__ == "__main__":
    test_full_fledged_demo()
    test_simplest_example()




def main():
    import yaml

    import pandas as pd
    import biocypher

    import ontoweaver

    bc = biocypher.BioCypher(
        biocypher_config_path = "test_2ontologies/biocypher.yaml",
        schema_config_path = "test_2ontologies/schema.yaml"
    )
    # bc.show_ontology_structure()

    table = pd.read_csv("oim.csv")

    with open("oim.yaml") as fd:
        mapping = yaml.full_load(fd)

    adapter = ontoweaver.tabular.extract_table(table, mapping)
    assert(adapter)

    assert(adapter.nodes)
    bc.write_nodes( adapter.nodes )

    assert(adapter.edges)
    bc.write_edges( adapter.edges )

    bc.write_import_call()

if __name__ == "__main__":
    main()



""" Interfaces for OntoWeaver.
"""
import logging
import math

import types as pytypes
import pandas as pd

from collections.abc import Iterable
from abc import ABCMeta as ABSTRACT, abstractmethod
from abc import abstractmethod as abstract
from typing import TypeAlias
from typing import Optional
from enum import Enum, EnumMeta

from . import errormanager
from . import serialize
from . import exceptions
from . import validate
from . import types as owtypes

logger = logging.getLogger("ontoweaver")


def is_not_null(val):
    """
    Checks if cell value is not empty nor 'null', 'nan'...

    Args:
        val: The value to check.

    Returns:
        bool: True if the value is valid, False otherwise.
    """
    if pd.api.types.is_numeric_dtype(type(val)):
        if (math.isnan(val) or val == float("nan")):
            return False
    elif str(val).lower() == "nan":  # Conversion from Pandas' `object` needs to be explicit.
        return False
    elif str(val) == "":
        return False
    elif str(val) == 'None':
        return False
    return True


class MetaEnum(EnumMeta):
    """
    Metaclass for Enum to allow checking if an item is in the Enum.
    """

    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        return True


class Enumerable(Enum, metaclass=MetaEnum):
    """
    Base class for Enums with MetaEnum metaclass.
    """
    pass


class TypeAffixes(str, Enumerable):
    """
    Enum for type affixes used in ID creation.
    """
    suffix = "suffix"
    prefix = "prefix"
    none = "none"


class ErrorManager:
    def __init__(self, raise_errors = True):
        self.raise_errors = raise_errors

    def error(self, msg, section = None, index = None, exception = exceptions.OntoWeaverError, indent = 0):

        location = ""
        if section:
            location = f" [for {section}"
            if index:
                location += f" #{index}"
            location += "]"

        err = "\t"*indent
        err += msg
        err += location

        if self.raise_errors:
            raise exception(err)
        else:
            logger.error(err)

        return err


class Element(metaclass = ABSTRACT):
    """Base class for either Node or Edge.

    Manages allowed properties mechanics."""

    def __init__(self,
                 id        : Optional[str] = None,
                 properties: Optional[dict[str,str]] = {},
                 label     : Optional[str] = None,
                 serializer: Optional[serialize.Serializer] = serialize.All(),
                 ):
        """Instantiate an element.

        :param str id: Unique identifier of the element. If id == None, is then set to the empty string.
        :param dict[str,str] properties: All available properties for this instance.
        :param str label: The label of the element.
        :param serializer: an instance indicating what to serialize to get a unique ID
        """
        if not id:
            self._id = ''
        else:
            self._id = str(id)

        # Use the setter to get sanity checks.
        self.properties = properties

        if not label:
            # Do not change the name here, or BioCypher will have problem
            # finding back labels.
            self._label = self.__class__.__name__ 
        else:
            self._label = str(label)

        self.serializer = serializer

    def __str__(self):
        return self.serializer(self)

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    @staticmethod
    @abstract
    def fields() -> list[str]:
        """List of property fields provided by the (sub)class."""
        raise NotImplementedError

    @abstract
    def as_tuple(self):
        """Convert the element class into Biocypher's expected tuple.

        Filter out properties along the way.
        """
        raise NotImplementedError

    @classmethod
    @abstract
    def from_tuple(cls,
                   biocypher_tuple : tuple,
                   serializer: Optional[serialize.Serializer] = serialize.All()
        ):
        # return cls(biocypher_tuple,serializer)
        raise NotImplementedError

    @property
    def id(self) -> str:
        return self._id

    @property
    def label(self) -> str:
        return self._label

    @property
    def properties(self) -> dict[str,str]:
        return self._properties

    @properties.setter
    def properties(self, properties: dict[str,str]):
        """Set available properties.

        Asserts that the passed properties are in the declared fields."""

        # Sanity checks:
        assert(properties is not None)
        assert(type(properties) == dict)
        # logger.debug(f"Properties of `{type(self).__name__}`: {list(properties.keys())}, available: {list(self.available())}")
        # TODO enable the usage of available() function to disable / enable parts of ontology / certain nodes
        # for p in properties:
        #     if p not in self.available():
        #         logger.error(f"\t\tProperty `{p}` should be available for type `{type(self).__name__}`, available ones: `{list(self.available())}`")
        #         assert(p in self.available())
        self._properties = properties


class Node(Element):
    """Base class for any Node."""

    def __init__(self,
                 id        : Optional[str] = None,
                 properties: Optional[dict[str,str]] = {},
                 label     : Optional[str] = None,  # Set from subclass name.
                 serializer: Optional[serialize.Serializer] = serialize.node.All(),
                 ):
        """Instantiate a Node.

        :param str id: Unique identifier of the node. If id == None, is then set to the empty string.
        :param dict[str,str] properties: All available properties for this instance.
        :param str label: The label of the node.
        :param serializer: an instance indicating what to serialize to get a unique ID
        """
        super().__init__(id = id, properties = properties, label = label, serializer = serializer)

    Tuple: TypeAlias = tuple[str,str,dict[str,str]]
    def as_tuple(self) -> Tuple:
        """Export the Node as a Biocypher tuple."""
        return (
            self._id,
            self._label,
            # FIXME this has been changed to keep ALL properties. No checking if allowed
            self.properties
        )

    @classmethod
    def from_tuple(cls,
                   biocypher_tuple : tuple[str,str,dict[str,str]],
                   serializer: Optional[serialize.Serializer] = serialize.node.All(),
                   ):
        assert(len(biocypher_tuple) == 3)
        return cls(
            id         = biocypher_tuple[0],
            label      = biocypher_tuple[1],
            properties = biocypher_tuple[2],
            serializer = serializer)

    def fields(self) -> list[str]:
        """List of property fields provided by the (sub)class."""
        return list(self.properties.keys())

    def __str__(self):
        return self.serializer(self)

    def __repr__(self):
        return f"({self.label}:{self.id}/{self.properties})"

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    # def serialize(self):
    #     return {
    #         "id": self._id,
    #         "label": self._label,
    #         "properties": self.properties,
    #         "serializer": self.serializer
    #     }

class Edge(Element):
    """Base class for any Edge."""

    def __init__(self,
                 id        : Optional[str] = None,
                 id_source : Optional[str] = None,
                 id_target : Optional[str] = None,
                 properties: Optional[dict[str,str]] = {},
                 label     : Optional[str] = None,  # Set from subclass name.
                 serializer: Optional[serialize.Serializer] = serialize.edge.All(),
                 ):
        """Instantiate an Edge.

        :param str id: Unique identifier of the edge. If id == None, is then set to the empty string.
        :param str id_source: Unique identifier of the source Node. If None, is then set to the empty string.
        :param str id_target: Unique identifier of the target Node. If None, is then set to the empty string.
        :param dict[str,str] properties: All available properties for this instance.
        :param str label: The label of the node.
        :param comparer: an instance indicating what to serialize to get a unique ID
        """
        super().__init__(id = id, properties = properties, label = label, serializer = serializer)
        self._id_source = str(id_source)
        self._id_target = str(id_target)

    @staticmethod
    @abstract
    def source_type():
        raise NotImplementedError

    @staticmethod
    @abstract
    def target_type():
        raise NotImplementedError

    @property
    def id_source(self):
        return self._id_source

    @id_source.setter
    def id_source(self, id_source):
        self._id_source = id_source

    @property
    def id_target(self):
        return self._id_target

    @id_target.setter
    def id_target(self, id_target):
        self._id_target = id_target

    Tuple: TypeAlias = tuple[str,str,str,dict[str,str]]
    def as_tuple(self) -> Tuple:
        """Export the Edge as a Biocypher tuple."""
        return (
            self._id,
            self._id_source,
            self._id_target,
            self._label,
            #.FIXME no checking if properties are allowed allowed_properties()
            self.properties
        )

    @classmethod
    def from_tuple(cls,
                   biocypher_tuple : tuple[str,str,str,str,dict[str,str]],
                   serializer: Optional[serialize.Serializer] = serialize.edge.All()
                   ):
        assert(len(biocypher_tuple) == 5)
        logging.debug(biocypher_tuple)
        return cls(
            id         = biocypher_tuple[0],
            id_source  = biocypher_tuple[1],
            id_target  = biocypher_tuple[2],
            properties = biocypher_tuple[4],
            label      = biocypher_tuple[3],
            serializer = serializer)

    def __repr__(self):
        if self.source_type() == Node:
            st = "."
        elif self.source_type() == None:
            st = ""
        else:
            st = f"{self.source_type().__name__}"
        if self.target_type() == Node:
            tt = "."
        else:
            tt = f"{'>'.join([t.__name__ for t in self.target_type()])}"

        return f"<({st}:{self.id_source})--[{self.label}:{self.id}/{self.properties}]-->({tt}:{self.id_target})>"

    def fields(self) -> list[str]:
        """List of property fields provided by the (sub)class."""
        return list(self.properties.keys())

    def __str__(self):
        return self.serializer(self)

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    # def serialize(self):
    #     return {
    #         "id": self._id,
    #         "id_source": self._id_source,
    #         "id_target": self._id_target,
    #         "label": self._label,
    #         "properties": self.properties,
    #         "serializer": self.serializer
    #     }

class GenericEdge(Edge):
    """Base class for any Edge."""

    def __init__(self,
                 id        : Optional[str] = None,
                 id_source : Optional[str] = None,
                 id_target : Optional[str] = None,
                 properties: Optional[dict[str,str]] = {},
                 label     : Optional[str] = None,  # Set from subclass name.
                 serializer: Optional[serialize.Serializer] = serialize.edge.All(),
                 ):
        """Instantiate an Edge.

        :param str id: Unique identifier of the edge. If id == None, is then set to the empty string.
        :param str id_source: Unique identifier of the source Node. If None, is then set to the empty string.
        :param str id_target: Unique identifier of the target Node. If None, is then set to the empty string.
        :param dict[str,str] properties: All available properties for this instance.
        :param str label: The label of the node.
        :param comparer: an instance indicating what to serialize to get a unique ID
        """
        super().__init__(id = id, id_source = id_source, id_target = id_target, properties = properties, label = label, serializer = serializer)

        logging.debug(f"GenericEdge ID: {id}")

    @staticmethod
    def source_type():
        return Node

    @staticmethod
    def target_type():
        return Node



class Adapter(errormanager.ErrorManager, metaclass = ABSTRACT):
    """Base class for implementing an adapter that consumes tabular data."""

    def __init__(self, raise_errors = True):
        """Allow to indicate which Element subclasses and which property fields
        are allowed to be exported by Biocypher.
        """
        self._nodes = []
        self._edges = []
        self.errors = []
        super().__init__(raise_errors)


    def nodes_append(self, node_s) -> None:
        """Append an Node (or each Node in a list of nodes) to the internal list of nodes."""
        if issubclass(type(node_s), Node):
            nodes = [node_s]
        else:
            nodes = node_s

        # logger.debug(f"Nodes: {nodes}.")
        for node in nodes:
            # logger.debug(f"\tAppend node {node}.")
            # Checking for duplicates in reconciliation, otherwise complexity too high.
            self._nodes.append(node.as_tuple())
            # return True

    def edges_append(self, edge_s) -> None:
        """Append an Edge (or each Edge in a list of edges) to the internal list of edges."""
        if issubclass(type(edge_s), Edge):
            edges = [edge_s]
        else:
            edges = edge_s

        # logger.debug(f"Edges: {edges}.")
        for edge in edges:
            # logger.debug(f"\tAppend edge {edge}.")
            # Checking for duplicates in reconciliation, otherwise complexity too high.
            self._edges.append(edge.as_tuple())
            # return True

    @property
    def nodes(self) -> Iterable[Node.Tuple]:
        """Return a generator yielding nodes."""
        for n in self._nodes:
            yield n

    @property
    def edges(self) -> Iterable[Edge.Tuple]:
        """Return a generator yielding edges."""
        for e in self._edges:
            yield e

    @abstractmethod
    def run(self):
        raise NotImplementedError()

    def __call__(self):
        for local_nodes, local_edges in self.run():
            yield local_nodes, local_edges


class Declare(errormanager.ErrorManager):
    """
    Declarations of functions used to declare and instantiate object classes used by the Adapter for the mapping
    of the data frame.

    Args:
        module: The module in which to insert the types declared by the configuration.
    """

    def __init__(self,
                 module=owtypes,
                 raise_errors = True,
                 ):
        super().__init__(raise_errors)
        self.module = module

        self.declared = []


    def make_node_class(self, name, properties={}, base=Node):
        """
        LabelMaker a node class with the given name and properties.

        Args:
            name: The name of the node class.
            properties (dict): The properties of the node class.
            base: The base class for the node class.

        Returns:
            The created node class.
        """
        # If type already exists, return it.
        if hasattr(self.module, name):
            cls = getattr(self.module, name)
            logger.debug(
                f"\t\tNode class `{name}` (prop: `{cls.fields()}`) already exists, I will not create another one.")
            for p in properties.values():
                if p not in cls.fields():
                    logger.warning(f"\t\t\tProperty `{p}` not found in declared fields for node class `{cls.__name__}`.")
            return cls

        def fields():
            return list(properties.values())

        attrs = {
            "__module__": self.module.__name__,
            "fields": staticmethod(fields),
        }
        t = pytypes.new_class(name, (base,), {}, lambda ns: ns.update(attrs))
        logger.debug(f"\t\tDeclare Node class `{t.__name__}` (prop: `{properties}`).")
        setattr(self.module, t.__name__, t)

        self.declared.append(t)
        return t


    def get_node_class(self, name):
        return getattr(self.module, name)


    def make_edge_class(self, name, source_t, target_t, properties={}, base=Edge):
        """
        LabelMaker an edge class with the given name, source type, target type, and properties.

        Args:
            name: The name of the edge class.
            source_t: The source type of the edge.
            target_t: The target type of the edge.
            properties (dict): The properties of the edge class.
            base: The base class for the edge class.

        Returns:
            The created edge class.
        """
        # If type already exists, check if the fields are the same.
        if hasattr(self.module, name):
            cls = getattr(self.module, name)
            cls_fields = cls.fields()

            # Compare the properties with the existing class fields
            if properties == cls_fields:
                logger.info(
                    f"\t\tEdge class `{name}` (prop: `{cls_fields}`) already exists with the same properties, I will not create another one.")
                return cls

            logger.warning(f"\t\tEdge class `{name}` already exists, but properties do not match.")
            # If properties do not match, we proceed to create a new class with the new properties.
            # FIXME: Would make much more sense to just append(?) new properties to existing class instead of creating new class.

        def fields():
            return properties

        def st():
            return source_t

        def tt():
            return [target_t]

        attrs = {
            "__module__": self.module.__name__,
            "fields": staticmethod(fields),
            "source_type": staticmethod(st),
            "target_type": staticmethod(tt),
        }
        t = pytypes.new_class(name, (base,), {}, lambda ns: ns.update(attrs))
        logger.debug(f"\t\tDeclare Edge class `{t.__name__}` (prop: `{properties}`).")
        setattr(self.module, t.__name__, t)

        self.declared.append(t)
        return t


    def get_edge_class(self, name):
        return getattr(self.module, name)


class MappingParser(Declare):
    # Various keys are allowed in the config to allow the user to use their favorite ontology vocabulary.
    k_row = ["row", "entry", "line", "subject", "source"]
    k_subject_type = ["to_subject", "to_object', 'to_node", "to_label", "to_type", "id_from_column", "id_from_element"]
    k_columns = ["columns", "fields", "column", "field", "element", "match_column", "id_from_column", "match_element", "id_from_element", "key", "keys"]
    k_target = ["to_target", "to_object", "to_node", "to_label", "to_type"]
    k_subject = ["from_subject", "from_source", "to_subject", "to_source", "to_node", "to_label", "to_type"]
    k_edge = ["via_edge", "via_relation", "via_predicate"]
    k_properties = ["to_properties", "to_property"]
    k_prop_to_object = ["for_objects", "for_object", "for_relation", "for_relations", "for_edge", "for_edges"]
    k_transformer = ["transformers"]
    k_metadata = ["metadata"]
    k_metadata_column = ["add_source_column_names_as"]
    k_validate_output = ["validate_output"]
    k_final_type = ["final_type", "final_object", "final_node", "final_subject", "final_label", "final_target"]
    k_reverse_edge = ["reverse_relation", "reverse_edge", "reverse_predicate", "reverse_link"]
    k_match_type_from = ["match_type_from_column", "match_type_from_element"]


class Transformer(errormanager.ErrorManager):
    """"Class used to manipulate cell values and return them in the correct format."""""

    def __init__(self,
        properties_of,
        value_maker = None,
        label_maker = None,
        branching_properties = None,
        columns = None,
        output_validator = None,
        multi_type_dict = None,
        raise_errors = True,
        **kwargs
    ):
        """
        Make a transformer class with the given parameters.
        Instantiate transformers.

        :param properties_of: the properties of each node type.
        :param value_maker: the ValueMaker object used for the logic of cell value selection for each transformer. Default is None.
        :param label_maker: the LabelMaker object used for handling the creation of the output of the transformer. Default is None.
        :param branching_properties: in case of branching on cell values, the dictionary holds the properties for each branch.
        :param columns: the columns to use in the mapping.
        :param output_validator: the OutputValidator object used for validating transformer output. Default is None.
        :param multi_type_dict: the dictionary holding regex patterns for node and edge type branching based on cell values.
        :param raise_errors: whether to raise errors or not. Default is True.
            each transformer is instantiated with a default OutputValidator object, and additional user-defined rules if needed in
            the tabular module.


        """
        super().__init__(raise_errors)

        self.properties_of = properties_of
        self.value_maker = value_maker
        self.label_maker = label_maker
        self.branching_properties = branching_properties
        self.columns = columns
        self.output_validator = output_validator
        if not self.output_validator:
            self.output_validator = validate.OutputValidator(validate.default_validation_rules, raise_errors = raise_errors)
        self.parameters = kwargs
        self.multi_type_dict = multi_type_dict
        self.final_type = None # The final type is to be passed by the label maker class based on the YAML mapping. That
                               # is why here it is None by default
        self.kwargs = kwargs
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.declare_types = Declare()



    def get_transformer(self):
        return self


    def __call__(self, row, i):
        """
        Process a row. If not empty/null, yields concatenated items as node IDs.

        Args:
            row: The current row of the DataFrame.
            i: The index of the current row.

        Yields:
            str: The concatenated string from the cell values.
        """
        for val in self.value_maker(self.columns, row, i):
            value, edge_type, node_type, reverse_edge = self.create(val, row)
            if is_not_null(value):
                yield value, edge_type, node_type, reverse_edge


    #FIXME: The functions below are never implemented.
    @abstractmethod
    def nodes(self):
        raise NotImplementedError

    @abstractmethod
    def edges(self):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def edge_type():
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def target_type():
       raise NotImplementedError

    @classmethod
    def source_type(cls):
       return cls.edge_type().source_type()

    def __repr__(self):

        representation = ""

        if hasattr(self, "from_subject"):
            from_subject = self.from_subject
        else:
            from_subject = "."

        target_name = ""
        edge_name = ""

        if hasattr(self, "multi_type_dict") and self.multi_type_dict is not None:
            for key, value in self.multi_type_dict.items():
                if value['to_object'] and value['via_relation']:
                    target_name = value['to_object']
                    edge_name = value['via_relation']
                elif value['to_object'] and not value['via_relation']:
                    target_name = value['to_object']
                    edge_name = "."


                if self.properties_of:
                    # The transformer is not a branching transformer, and has only one set of properties.
                    props = "{"
                    trans = []
                    for kind in self.properties_of:
                        trans.append(f"{kind} {self.properties_of[kind]}>")
                    props += ", ".join(trans)
                    props+="}"
                elif self.branching_properties and self.branching_properties.get(value['to_object'], None):
                    # The transformer is a branching transformer, and has multiple sets of properties. We extract the ones for the current type.
                    props = self.branching_properties.get(value['to_object'])
                else:
                    # The transformer has no properties for the type.
                    props = "{}"


                params = ""
                parameters = {k:v for k,v in self.parameters.items() if k not in ['subclass', 'from_subject', "match"]}
                if parameters:
                    p = []
                    for k,v in parameters.items():
                        p.append(f"{k}={v}")
                    params = ','.join(p)

                if from_subject == "." and edge_name == "." and target_name == "." and props == "{}":
                    # If this is a property transformer
                    link = ""

                elif from_subject == "." and edge_name == "." and (target_name != "." or props != "{}"):
                    # This a subject transformer.
                    link = f" => ({target_name}/{props})"

                else:
                    # This is a regular transformer.
                    link = f" => ({from_subject})--[{edge_name.__name__}]->({target_name.__name__}/{props})"

                if self.columns:
                    columns = self.columns
                else:
                    columns = []

                # for c in columns:
                    # if not isinstance(c, str):
                    #     self.error(f"Column `{c}` is not a string, did you mistype a leading colon?", exception=exceptions.ParsingError)
                columns = [str(i) for i in columns]

                representation += (f"<{type(self).__name__}({params}) {','.join(columns)}{link}>")

        else:
            #The transformer is a property transformer. We add the property name and value in the tabular.properties() function.
            representation += (f"<{type(self).__name__}() {','.join(self.columns) if self.columns else ''} =>")

        return representation


    def validate(self, res):
        """
        Validate the output of the transformer, using the output_validator. of the transformer instance.
        """
        try:
            if isinstance(self.output_validator, validate.SimpleOutputValidator) or isinstance(self.output_validator, validate.SkipValidator):
                # The SimpleOutputValidator and SkipValidator do not use the Pandera package, and the value is therefore not
                # required to be in a DataFrame format. Voiding the creation of a DataFrame here saves computational time for
                # large datasets.
                if self.output_validator(res):
                    return True
                else:
                    return False
            else:
                if self.output_validator(pd.DataFrame([res], columns=["cell_value"])):
                    return True
                else:
                    return False
        except pa.errors.SchemaErrors as error:
            msg = f"Transformer {self.__repr__()} did not produce valid data {error}."
            self.error(msg, exception = exceptions.DataValidationError)


    def create(self, returned_value, row):
        """
        Create the output of the transformer, using the label_maker of the transformer instance.

        Returns:
            - Extracted cell value (can be node ID, property value, edge ID),
            - edge type,
            - target node type, and
            - reverse relation in case declared in the mapping.
        """
        result_object = self.label_maker(self.validate, returned_value, self.multi_type_dict, self.branching_properties, row)
        if not result_object:
            return None, None, None, None
        if result_object.target_node_type:
            self.target_type = result_object.target_node_type.__name__
        if result_object.target_element_properties is not None:
            self.properties_of = result_object.target_element_properties
        self.final_type = result_object.final_type
        return result_object.extracted_cell_value, result_object.edge_type, result_object.target_node_type, result_object.reverse_relation


class All:
    """Gathers lists of subclasses of Element and their fields
    existing in a given module.

    Is generally used to label_maker an `all` variable in a module:
    .. code-block:: python

        all = base.All(sys.modules[__name__])

    Which can later be used to pass all available Element types to an Adapter:
    .. code-block:: python

        a = MyAdapter( node_types = MyModule.all.nodes() )
    """

    def __init__(self, module):
        self.module = module

    def elements(self, asked: Element = Element) -> list[Element]:
        m = self.module.__dict__
        classes = []
        for c in m:
            if isinstance(m[c], type) \
            and m[c].__module__ == self.module.__name__ \
            and issubclass(m[c], asked):
                classes.append(m[c])
                logger.debug(f"Found `{asked.__name__}` class: `{m[c]}` (prop: `{m[c].fields()}`).")
                # t = m[c]
                # logger.debug(f"\t\t#### {t.mro()[:-3]}/{t.__name__} => {t.fields()}")
        return classes

    def nodes(self) -> list[Node]:
        return self.elements(Node)

    def edges(self) -> list[Edge]:
        return self.elements(Edge)

    def node_fields(self) -> list[str]:
        names = [] # FIXME use a set?
        for c in self.nodes():
            names += c.fields()
        return names

    def edge_fields(self) -> list[str]:
        names = [] # FIXME use a set?
        for c in self.edges():
            names += c.fields()
        return names




import logging
import io
import yaml
import pandas as pd

import ontoweaver

class simplest(ontoweaver.base.Transformer):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.declare_types.make_node_class("patient")
        self.declare_types.make_node_class("variant")

        self.declare_types.make_edge_class("patient_has_variant", "patient", "variant")

    def __call__(self, row, i):
        yield self.create(str(row["patient"]), row)


def test_transformer_simplest():
    # Add the passed transformer to the list available to OntoWeaver.
    ontoweaver.transformer.register(simplest)

    logging.debug("Load mapping...")
    yaml_mapping = """
    row:
        map:
            column: variant
            to_subject: variant
    transformers:
        - simplest:
            to_object: patient
            via_relation: patient_has_variant
    """

    mapping = yaml.safe_load(yaml_mapping)

    logging.debug("Load data...")
    data = """variant,patient
A,P1
B,P2
"""
    table = pd.read_csv(io.StringIO(data))

    logging.debug("Run the adapter...")
    adapter = ontoweaver.extract_table(table, mapping)


if __name__ == "__main__":
    test_transformer_iftypes()




def test_custom_transformer():
    from . import testing_functions
    import logging
    import ontoweaver
    from tests.custom_transformer.custom import OmniPath

    logging.basicConfig(level=logging.DEBUG)

    directory_name = "custom_transformer"

    expected_nodes = [
        ('Q9JMA7:protein', 'protein', {'genesymbol': 'Cyp3a41a; Cyp3a41b', 'ncbi_tax_id': '10090'}),
        ('P48281:protein', 'protein', {'genesymbol': 'Vdr', 'ncbi_tax_id': '10090'})
    ]

    expected_edges = [
        ('(P48281:protein)--[transcriptional]->(Q9JMA7:protein)', 'P48281:protein', 'Q9JMA7:protein', 'transcriptional',
            {'is_directed': 'True', 'is_stimulation': 'True', 'is_inhibition': 'False', 'consensus_direction': 'True', 'consensus_stimulation': 'True', 'consensus_inhibition': 'False', 'sources': 'ExTRI_CollecTRI;HTRI_CollecTRI;HTRIdb;HTRIdb_DoRothEA;NTNU.Curated_CollecTRI;SIGNOR_CollecTRI;TRRUST;TRRUST_CollecTRI;TRRUST_DoRothEA', 'references': 'CollecTRI:11723248;CollecTRI:11991950;CollecTRI:12147248;', 'omnipath': 'False', 'kinaseextra': 'False', 'ligrecextra': 'False', 'pathwayextra': 'False', 'mirnatarget': 'False', 'dorothea': 'True', 'collectri': 'True', 'tf_target': 'True', 'lncrna_mrna': 'False', 'tf_mirna': 'False', 'small_molecule': 'False', 'dorothea_curated': 'True', 'dorothea_chipseq': 'False', 'dorothea_tfbs': 'False', 'dorothea_coexp': 'False', 'dorothea_level': 'A;D', 'type': 'transcriptional', 'curation_effort': '21', 'extra_attrs': 'blabla', 'evidences': 'blabli'})
    ]

    ontoweaver.transformer.register(OmniPath)

    data_mapping = {f"tests/{directory_name}/data.tsv": f"tests/{directory_name}/mapping.yaml"}

    nodes, edges = ontoweaver.extract(data_mapping, affix="suffix", raise_errors=False, sep = "\t")

    fnodes, fedges = ontoweaver.fusion.reconciliate(ontoweaver.ow2bc(nodes), ontoweaver.ow2bc(edges))

    logging.debug(fnodes)
    logging.debug(fedges)
    testing_functions.assert_equals(fnodes, expected_nodes)
    testing_functions.assert_equals(fedges, expected_edges)


if __name__ == "__main__":
    test_custom_transformer()



""" Information fusion features for finding duplicates.
"""
import logging
from abc import ABCMeta as ABSTRACT, ABCMeta, abstractmethod

from . import base
from . import serialize

logger = logging.getLogger("ontoweaver")


class Congregater(metaclass=ABCMeta):
    """Interface for classes detecting duplicated elements (nodes or edges) in BioCypher's tuple lists.

    Their contract is to produce a dictionary mapping a key to a list of duplicates.
    The key takes the form of a string made from serializing some members of the targeted elements.
    The serialization is delegated to an object of type serialize.Serialize.

    The result of the duplicates congregation is available through the `duplicates` property.

    Derived classes should implement the __call__ method.
    """

    def __init__(self, serializer: serialize.Serializer = serialize.All()):
        self._serializer = serializer
        self._duplicates = {}

    def __len__(self):
        return len(self._duplicates)

    @property
    def serializer(self):
        return self._serializer

    @property
    def duplicates(self):
        return self._duplicates

    @abstractmethod
    def __call__(self, biocypher_tuples):
        """Call interface

        Args:
            biocypher_tuples: a list of tuples in the BioCypher format for nodes and/or edges.
        """
        raise NotImplementedError


class Congregate(Congregater):
    """A Congregater that detects duplicated elements of the given type (derivatirng from base.Element)."""

    def __init__(self, elem_cls: base.Element, serializer: serialize.Serializer = serialize.All()):
        """ Constructor.

        Args:
            elem_cls: the class of the Elements that will be processed.
            serializer: a serialize.Serializer object giving the key on which to detect duplicated elements.
        """
        logger.debug(f"Instantiate Congregate {type(self).__name__} for element {elem_cls.__name__} with serializer {type(serializer).__name__}")
        assert(issubclass(elem_cls, base.Element))
        self._elem_cls = elem_cls
        super().__init__(serializer)

    def __call__(self, biocypher_tuples):
        """Call interface

        Args:
            biocypher_tuples: a list of tuples in the BioCypher format for nodes xor edges.
        """
        logger.debug("Call Congregate...")
        for t in biocypher_tuples:
            elem = self._elem_cls.from_tuple(t, serializer = self.serializer)
            self._duplicates[elem] = self._duplicates.get(elem, []) + [elem]
            yield elem

        if logger.level == "DEBUG":
            logger.debug(f"Congregated in {len(self._duplicates)} keys:")
            for k,l in self._duplicates.items():
                logger.debug(f"  Key `{k}` => {len(l)} elements")


class Nodes(Congregate):
    """A Congregater that detects duplicated Nodes."""

    def __init__(self, serializer: serialize.Serializer = serialize.All()):
        """ Constructor.

        Args:
            serializer: a serialize.Serializer object giving the key on which to detect duplicated Nodes.
        """
        super().__init__(base.Node, serializer)


class Edges(Congregate):
    """A Congregater that detects duplicated Edges."""

    def __init__(self, serializer: serialize.Serializer = serialize.All()):
        """ Constructor.

        Args:
            serializer: a serialize.Serializer object giving the key on which to detect duplicated Nodes.
        """
        super().__init__(base.GenericEdge, serializer)




""" Classes related to loading input data.
"""
import glob
import rdflib
import pathlib
import logging
import json as pyjson
import pandas as pd
from abc import ABCMeta as ABSTRACT, abstractmethod

from . import tabular
from . import owl
from . import xml
from . import json
from . import exceptions

logger = logging.getLogger("ontoweaver")


class Loader(metaclass = ABSTRACT):
    """ The interface for classes that load data.

        Classes implementing this interface are in charge of checking if they can
        load the given data, and indicate which base.Adapter can manage what they load.
    """
    def __call__(self, data, **kwargs):
        """If the loader allows the passed data, then load and return them.

           This will pass any named argument to the load() function.

           Args:
               data: the data to be loaded.
               kwargs: any named argument dictionary

           Returns:
               The loaded data structure.
        """
        if self.allows(data):
            return self.load(data, **kwargs)

    @abstractmethod
    def allows(self, data):
        """ Returns True if the given data can be loaded by this Loader.

            The given data can be string refereing to a filename,
            or any typed data structure.

            Args:
                data: the data to be checked

            Returns:
                True, if the given data is allowed, else False.
        """
        raise NotImplementedError()

    @abstractmethod
    def load(self, data : list, **kwargs):
        """ Load the data structure from the given item.

            The given item can be a string representing a filename,
            or a typed data structure. In the later case, this method
            will probably just return it as-is.

            Args:
                data: the list of data items to load
                kwargs: any named argument that this method may use or pass to subfunctions
        """
        raise NotImplementedError()

    @abstractmethod
    def adapter(self, **kwargs):
        """ Return the type of the adapter that can process the loaded data.

            Args:
                kwargs: any named argument that this method may use or pass
        """
        return NotImplementedError()

    def extensions(self, filenames):
        """ Return the extensions of the given filenames

            Utility function for loaders interested in filenames.

            Args:
                filenames: a list of filenames (strings or pathlid.Path objects)

            Returns:
                A list of extensions, with their dot.
        """
        assert type(filenames) == list, "You must pass a list to this function"

        if not all(type(f) == str for f in filenames):
            # Those cannot be extensions
            return None

        extensions = []
        for filename in filenames:
            if type(filename) == str or type(filename) == pathlib.Path:
                ext = pathlib.Path(filename).suffix
                if not ext:
                    # If the passed filename is itself an extension,
                    # just use it as-is.
                    ext = filename
                assert ext != '', f"I can't parse the extension of file {filename}"
                extensions.append(ext)
            else:
                logger.warning(f"I don't know how to handle the filename `{filename}` of type `{type(filename)}`. I'll pretend I saw nothing, but this may generate errors later on.")
        return extensions


class LoadPandasDataframe(Loader):
    """ Load Pandas DataFrame

        This mainly concatenates the given DataFrames into a single one.
        Pandas will raises an error if that's not possible.
    """
    def allows(self, data):
        return all(type(d) == pd.DataFrame for d in data)

    def load(self, dataframes, **kwargs):
        logger.debug(dataframes)
        assert type(dataframes) != str and len(dataframes) > 0 and not any(d.empty for d in dataframes), "A Loader expects a list (or an iterable) of data."
        return pd.concat(dataframes)

    def adapter(self, **kwargs):
        return tabular.PandasAdapter


class LoadPandasFile(Loader):
    """Read a file with Pandas, using its extension to guess its format.

    If no additional arguments are passed, it will call the
    Pandas `read_*` function with `filter_na = False`, which makes empty cell
    values to be loaded as empty strings instead of NaN values.

    Args:
        filename: The name of the data file the user wants to map.
        separator (str): The separator used in the data file. Defaults to None.
        kwargs: A dictionary of arguments to pass to pandas.read_* functions.

    Raises:
        exceptions.FeatureError: if the extension is unknown.

    Returns:
        A Pandas DataFrame.
    """

    def __init__(self):
            # We probably don't want NaN as a default,
            # since they tend to end up in a label.
            #'c' engine does not support regex separators (separators > 1 char and different
            # from '\s+' are interpreted as regex) which results in an error.
        self.read_funcs = {
            '.csv'    : (pd.read_csv, {
                'sep': ',',
                'na_filter': True, # Do not load empty cells, if possible.
                'dtype': str, # Always load data as string, to avoid conversion of numbers to floating-point values with decimal separators.
                'engine': 'python'
            }),
            '.tsv'    : (pd.read_csv, {
                'sep': '\t',
                'na_filter': True,
                'dtype': str,
                'engine': 'python'
            }),
            '.txt'    : (pd.read_csv, {
                'na_filter': True,
                'dtype': str,
                'engine': 'python'
            }),
            '.dat'    : (pd.read_csv, {
                'na_filter': True,
                'dtype': str,
                'engine': 'python'
            }),

            # Use dtype: str for formats not supposed to host complex objects in their cells.
            '.xls'    : (pd.read_excel,   {'na_filter': True, 'dtype': str}),
            '.xlsx'   : (pd.read_excel,   {'na_filter': True, 'dtype': str}),
            '.xlsm'   : (pd.read_excel,   {'na_filter': True, 'dtype': str}),
            '.xlsb'   : (pd.read_excel,   {'na_filter': True, 'dtype': str}),
            '.odf'    : (pd.read_excel,   {'na_filter': True, 'dtype': str}),
            '.ods'    : (pd.read_excel,   {'na_filter': True, 'dtype': str}),
            '.odt'    : (pd.read_excel,   {'na_filter': True, 'dtype': str}),
            '.orc'    : (pd.read_orc,     {'na_filter': True, 'dtype': str}),
            '.sas'    : (pd.read_sas,     {'na_filter': True, 'dtype': str}),
            '.spss'   : (pd.read_spss,    {'na_filter': True, 'dtype': str}),
            '.stata'  : (pd.read_stata,   {'na_filter': True, 'dtype': str}),

            # Allow multiple dtypes for the others.
            '.json'   : (pd.read_json,    {'na_filter': True}),
            '.feather': (pd.read_feather, {'na_filter': True}),
            '.pickle' : (pd.read_pickle,  {'na_filter': True}),
            '.hdf'    : (pd.read_hdf,     {'na_filter': True}),
            '.parquet': (pd.read_parquet, {}), # There's no na_filter in read_parquet.
        }


    def allows(self, filenames):
        # assert type(filenames) != str and len(filenames) > 0 and all(i != '' for i in filenames), "A Loader expects a list (or an iterable) of data."

        exts = self.extensions(filenames)
        if not exts:
            return False

        if all(e in self.read_funcs.keys() for e in exts):
            return True
        else:
            return False


    def load(self, filenames, **kwargs):
        assert type(filenames) != str and len(filenames) > 0 and all(i != '' for i in filenames), "A Loader expects a list (or an iterable) of data."

        exts = list(set(self.extensions(filenames)))
        assert exts
        if not self.allows(exts):
            msg = f"One of those file formats: `{', '.join(exts)}` is not supported (I can only read one of: {', '.join(self.read_funcs.keys())})"
            logger.error(msg)
            raise exceptions.FeatureError(msg)

        expanded = []
        for filename in filenames:
            globbed = glob.glob(filename)
            expanded += globbed
        filenames = expanded

        data = []
        for filename in filenames:
            loadfunc  = self.read_funcs[pathlib.Path(filename).suffix][0]
            kw = self.read_funcs[pathlib.Path(filename).suffix][1]

            # Overwrite default named arguments with the passed ones.
            kw.update(kwargs)

            if not kw:
                # pd.read_parquet does not allow even an empty kwargs.
                logger.debug(f"No argument passed to `{pathlib.Path(filename).suffix}` load function.")
                try:
                    data.append( loadfunc(filename) )
                except AttributeError as e:
                    logger.error(f"I cannot load those parquet files. Maybe they don't have the 'name' property? That may explain the following error in Pandas.")
                    raise e
            else:
                logger.debug(f"Additional arguments passed to the `{pathlib.Path(filename).suffix}` load function: {kw}")
                data.append( loadfunc(filename, **kw) )

        out = pd.concat(data)
        return out

    def adapter(self, **kwargs):
        return tabular.PandasAdapter


class LoadOWLGraph(Loader):
    """ Load the given rdflib.Graph(s)

        If adapter() function is given the string "automap", this will select the
        owl.OWLAutoAdapter, else the owl.OWLAdapter.

        If you try to load several graphs, this will show a warning stating
        that it is quite improbable that this would work.
    """
    def allows(self, data):
        return all(type(d) == rdflib.Graph for d in data)

    def load(self, graphs, **kwargs):
        assert type(graphs) != str and len(graphs) > 0 and all(i != '' for i in graphs), "A Loader expects a list (or an iterable) of data."
        if len(graphs) > 1:
            logger.warning(
                "Loading multiple OWL graphs at once does not guarantee a correct fusion." \
                " That is, graph operations in RDFLib are assumed to be performed in subgraphs" \
                " of some larger database and assume shared blank node IDs, and therefore may" \
                " cause unwanted collisions of blank-nodes in graph (cf." \
                " https://rdflib.readthedocs.io/en/7.1.1/merging.html)." \
                " To avoid this warning, either load only compatible OWL graphs at once" \
                " either, load OWL graphs separately and" \
                " use OntoWeaver's fusion features afterward.")

        graph = rdflib.Graph()
        for g in graphs:
            graph += g

        return graph

    def adapter(self, **kwargs):
        if "automap" in kwargs:
            logger.debug("Asked for `automap`, I'll use owl.OWLAutoAdapter")
            return owl.OWLAutoAdapter
        else:
            logger.debug("Asked for regular mapping, I'll use owl.OWLAdapter")
            return owl.OWLAdapter


class LoadOWLFile(Loader):
    """ Load the given OWL file.

        If adapter() function is given the string "automap", this will select the
        owl.OWLAutoAdapter, else the owl.OWLAdapter.

        If you try to load several graphs, this will show a warning stating
        that it is quite improbable that this would work.
    """
    def __init__(self):
        self.allowed = [".owl", ".n3", ".turtle", ".ttl", ".nt", ".trig", ".trix", ".json-ld"]

    def allows(self, filenames):
        # assert type(filenames) != str and len(filenames) > 0 and all(i != '' for i in filenames), "A Loader expects a list (or an iterable) of data."

        exts = self.extensions(filenames)
        if not exts:
            return False

        if all(e in self.allowed for e in exts):
            return True
        else:
            return False

    def load(self, filenames, **kwargs):
        assert type(filenames) != str and len(filenames) > 0 and all(i != '' for i in filenames), "A Loader expects a list (or an iterable) of data."
        if len(filenames) > 1:
            logger.warning(
                "Loading multiple OWL files at once does not guarantee a correct fusion." \
                " That is, graph operations in RDFLib are assumed to be performed in subgraphs" \
                " of some larger database and assume shared blank node IDs, and therefore may" \
                " cause unwanted collisions of blank-nodes in graph (cf." \
                " https://rdflib.readthedocs.io/en/7.1.1/merging.html)." \
                " To avoid this warning, either load only compatible OWL files at once" \
                " either, load OWL files separately and" \
                " use OntoWeaver's fusion features afterward.")

        exts = set(self.extensions(filenames))
        assert exts
        for ext in exts:
            if not self.allows([ext]):
                msg = f"File format '{ext}' is not supported (I can only read one of: {' ,'.join(self.read_funcs.keys())})"
                logger.error(msg)
                raise exceptions.FeatureError(msg)

        g = rdflib.Graph()
        for filename in filenames:
            ext = pathlib.Path(filename).suffix
            if ext == ".owl":
                g.parse(filename, format = "xml")
            else:
                g.parse(filename) # Guess the format based on extension.

        return g


    def adapter(self, **kwargs):
        if "automap" in kwargs:
            logger.debug("Asked for `automap`, I'll use owl.OWLAutoAdapter")
            return owl.OWLAutoAdapter
        else:
            logger.debug("Asked for regular mapping, I'll use owl.OWLAdapter")
            return owl.OWLAdapter


class LoadXMLString(Loader):
    """ Load (here, that is: parse) the given XML string.

        If you try to load several graphs, this will show a warning stating
        that it is quite improbable that this would work.
    """
    def allows(self, data):
        # assert type(data) != str and len(data) > 0 and all(i != '' for i in data), "A Loader expects a list (or an iterable) of data."
        return all(type(d) == str for d in data)

    def load(self, xmls, **kwargs):
        assert type(xmls) != str and len(xmls) > 0 and all(i != '' for i in xmls), "A Loader expects a list (or an iterable) of xmls."
        if len(xmls) > 1:
            logger.warning(
                "Loading multiple XML files at once is not really supported," \
                " since I'm only going to concatenate the documents." \
                " I thus cannot guarantee a well-formed XML document at the end." \
                " Ignore this warning only if you know how to sequentially assemble a valid XML document.")

        return "\n".join(xmls)

    def adapter(self, **kwargs):
        return xml.XMLAdapter


class LoadXMLFile(Loader):
    """ Load the given XML file.

        If you try to load several graphs, this will show a warning stating
        that it is quite improbable that this would work.
    """
    def __init__(self):
        self.xl = LoadXMLString()
        self.allowed = [".xml"]

    def allows(self, filenames):
        # assert type(data) != str and len(data) > 0 and all(i != '' for i in data), "A Loader expects a list (or an iterable) of data."

        exts = self.extensions(filenames)
        if not exts:
            return False

        if all(e in self.allowed for e in exts):
            return True
        else:
            return False

    def load(self, filenames, **kwargs):
        assert type(filenames) != str and len(filenames) > 0 and all(i != '' for i in filenames), "A Loader expects a list (or an iterable) of filenames."
        if len(filenames) > 1:
            logger.warning(
                "Loading multiple XML files at once is not really supported," \
                " since I'm only going to concatenate the documents." \
                " I thus cannot guarantee a well-formed XML document at the end." \
                " Ignore this warning only if you know how to sequentially assemble a valid XML document.")

        data = []
        for filename in filenames:
            ext = pathlib.Path(filename).suffix
            if not self.allows([filename]):
                msg = f"File format '{ext}' of file '{filename}' is not supported (I can only read one of: {' ,'.join(self.allowed)})"
                logger.error(msg)
                raise exceptions.FeatureError(msg)

            with open(filename, 'r') as fd:
                data.append( fd.read() )

        return self.xl(data)

    def adapter(self, **kwargs):
        return xml.XMLAdapter


class LoadJSONString(Loader):
    """ Load (here, that is: parse) the given JSON string.
    """
    def allows(self, jsons):
        # assert type(jsons) != str and len(jsons) > 0 and all(i != '' for i in jsons), "A Loader expects a list (or an iterable) of data."
        if not all(type(j) == str for j in jsons):
            return False

        for j in jsons:
            try:
                pyjson.loads(j)
            except ValueError as e:
                return False

        return True

    def load(self, jsons, **kwargs):
        dic = {}
        logger.debug("Assembling JSON strings into a single data structure.")
        for j in jsons:
            d = pyjson.loads(j)
            for k in d:
                if k in dic:
                    logger.warning(
                        f"Overwriting `{k}:{dic[k]}` with `{d[k]}` in a previously loaded JSON." \
                        " This happens if you ask to load multiple non-independant JSONs." \
                        " You may also want to double-check that the JSON list has the order that you want." \
                    )
            dic.update(d)
        return pyjson.dumps(dic)

    def adapter(self, **kwargs):
        return json.JSONAdapter


class LoadJSONFile(Loader):
    """ Load the given JSON file.
    """
    def __init__(self):
        self.allowed = [".json"]
        self.jl = LoadJSONString()

    def allows(self, filenames):
        # assert type(filenames) != str and len(filenames) > 0 and all(i != '' for i in filenames), "A Loader expects a list (or an iterable) of data."
        return all(pathlib.Path(f).suffix in self.allowed for f in filenames)

    def load(self, filenames, **kwargs):
        assert type(filenames) != str and len(filenames) > 0 and all(i != '' for i in filenames), "A Loader expects a list (or an iterable) of data."

        jsons = []
        for filename in filenames:
            ext = pathlib.Path(filename).suffix
            if not self.allows([filename]):
                msg = f"File format '{ext}' of file '{filename}' is not supported (I can only read one of: {' ,'.join(self.allowed)})"
                logger.error(msg)
                raise exceptions.FeatureError(msg)

            with open(filename, 'r') as fd:
                jsons.append( fd.read() )

        return self.jl.load(jsons)

    def adapter(self, **kwargs):
        return json.JSONAdapter

